from cms.models import Page
from cms.models.pluginmodel import CMSPlugin
from cms.plugin_pool import plugin_pool

from .base import SubcommandsCommand


class ListApphooksCommand(SubcommandsCommand):
    help_string = 'Lists all apphooks in pages'
    command_name = 'apphooks'

    def handle(self, *args, **options):
        urls = list(
            Page.objects.exclude(application_urls='').exclude(
                application_urls__isnull=True
            ).values_list(
                'application_urls', 'application_namespace'
            )
        )
        apphooks = {}
        for apphook, application_namespace in urls:
            if apphook in apphooks:
                apphooks[apphook][0].append('active')
            else:
                apphooks[apphook] = [['active'], application_namespace]
        for apphook, attributes in apphooks.items():
            attributes[0].sort()
            if attributes[1]:
                self.stdout.write('{0}[instance: {1}] ({2})\n'.format(
                    apphook, attributes[1], '/'.join(attributes[0])
                ))
            else:
                self.stdout.write('{0} ({1})\n'.format(
                    apphook, '/'.join(attributes[0])
                ))


def plugin_report():
    """
    Returns a report of existing plugins

    structure of report:
    [
        {
            'type': CMSPlugin class,
            'model': plugin_type.model,
            'instances': instances in the CMSPlugin table,
            'unsaved_instances': those with no corresponding model instance,
        },
    ]
    """
    plugin_report = []
    all_plugins = CMSPlugin.objects.order_by('plugin_type')
    plugin_types = list(set(all_plugins.values_list('plugin_type', flat=True)))
    plugin_types.sort()

    for plugin_type in plugin_types:
        plugin = {}
        plugin['type'] = plugin_type
        try:
            # get all plugins of this type
            plugins = CMSPlugin.objects.filter(plugin_type=plugin_type)
            plugin['instances'] = plugins
            # does this plugin have a model? report unsaved instances
            plugin['model'] = plugin_pool.get_plugin(name=plugin_type).model
            unsaved_instances = [p for p in plugins if not p.get_plugin_instance()[0]]
            plugin['unsaved_instances'] = unsaved_instances

        # catch uninstalled plugins
        except KeyError:
            plugin['model'] = None
            plugin['instances'] = plugins
            plugin['unsaved_instances'] = []

        plugin_report.append(plugin)

    return plugin_report


class ListPluginsCommand(SubcommandsCommand):
    help_string = 'Lists all plugins in CMSPlugin'
    command_name = 'plugins'

    def handle(self, *args, **options):
        self.stdout.write('==== Plugin report ==== \n\n')
        self.stdout.write('There are %s plugin types in your database \n' % len(plugin_report()))
        for plugin in plugin_report():
            self.stdout.write('\n%s \n' % plugin['type'])

            plugin_model = plugin['model']
            instances = len(plugin['instances'])
            unsaved_instances = len(plugin['unsaved_instances'])

            if not plugin_model:
                self.stdout.write(self.style.ERROR('  ERROR      : not installed \n'))

            elif plugin_model == CMSPlugin:
                self.stdout.write('    model-less plugin \n')
                self.stdout.write('    unsaved instance(s) : %s  \n' % unsaved_instances)

            else:
                self.stdout.write('  model      : %s.%s  \n' % (plugin_model.__module__, plugin_model.__name__))
                if unsaved_instances:
                    self.stdout.write(self.style.ERROR('  ERROR      : %s unsaved instance(s) \n' % unsaved_instances))

            self.stdout.write('  instance(s): %s \n' % instances)


class ListCommand(SubcommandsCommand):
    help_string = 'List objects of the following types:'
    command_name = 'list'
    subcommands = {
        'apphooks': ListApphooksCommand,
        'plugins': ListPluginsCommand
    }



import uuid

from cms.models import Page
from cms.signals import post_obj_operation, pre_obj_operation


def send_pre_page_operation(request, operation, sender=Page, **kwargs):
    token = str(uuid.uuid4())
    pre_obj_operation.send(
        sender=sender,
        operation=operation,
        request=request,
        token=token,
        **kwargs
    )
    return token


def send_post_page_operation(request, operation, token, sender=Page, **kwargs):
    post_obj_operation.send(
        sender=sender,
        operation=operation,
        request=request,
        token=token,
        **kwargs
    )



from django.contrib import admin
from django.utils.translation import gettext as _

from cms.plugin_base import CMSPluginBase
from cms.plugin_pool import plugin_pool
from cms.test_utils.project.pluginapp.plugins.manytomany_rel.models import (
    Article,
    ArticlePluginModel,
    PluginModelWithFKFromModel,
    PluginModelWithM2MToModel,
)


class ArticlePlugin(CMSPluginBase):
    model = ArticlePluginModel
    name = _("Articles")
    render_template = "articles.html"
    admin_preview = False
    filter_horizontal = ['sections']

    def render(self, context, instance, placeholder):
        article_qs = Article.objects.filter(section__in=instance.sections.all())
        context.update({'instance': instance,
                        'article_qs': article_qs,
                        'placeholder': placeholder})
        return context


class ArticlePluginAdmin(admin.ModelAdmin):
    filter_horizontal = ['sections']


class ArticleDynamicTemplatePlugin(CMSPluginBase):
    model = ArticlePluginModel
    name = _("Articles")
    admin_preview = False

    def get_render_template(self, context, instance, placeholder):
        if instance.title == 'custom template':
            return "articles_custom.html"
        else:
            return "articles.html"

    def render(self, context, instance, placeholder):
        article_qs = Article.objects.filter(section__in=instance.sections.all())
        context.update({'instance': instance,
                        'article_qs': article_qs,
                        'placeholder': placeholder})
        return context

###


class PluginWithFKFromModel(CMSPluginBase):
    model = PluginModelWithFKFromModel
    render_template = "articles.html"


class PluginWithM2MToModel(CMSPluginBase):
    model = PluginModelWithM2MToModel
    render_template = "articles.html"


plugin_pool.register_plugin(ArticlePlugin)
plugin_pool.register_plugin(ArticleDynamicTemplatePlugin)
plugin_pool.register_plugin(PluginWithM2MToModel)
plugin_pool.register_plugin(PluginWithFKFromModel)

# Used to test integrity of plugin form
admin.site.register(ArticlePluginModel, ArticlePluginAdmin)



from cms.plugin_base import CMSPluginBase
from cms.plugin_pool import plugin_pool


class PlanetPlugin(CMSPluginBase):
    """
    The contents of this plugin are never rendered.
    Only render the toolbar entries.
    """
    parent_classes = ['SolarSystemPlugin']
    render_template = 'nested_plugins_app/planet.html'
    require_parent = True


class SolarSystemPlugin(CMSPluginBase):
    allow_children = True
    render_template = 'nested_plugins_app/solar_system.html'
    child_classes = ['PlanetPlugin']


plugin_pool.register_plugin(PlanetPlugin)
plugin_pool.register_plugin(SolarSystemPlugin)



from django.contrib import admin
from django.contrib.admin import ModelAdmin

from cms.test_utils.project.pluginapp.plugins.manytomany_rel.models import (
    Article,
    Section,
)

admin.site.register(Section, ModelAdmin)
admin.site.register(Article, ModelAdmin)



from django.forms.models import model_to_dict
from django.test.utils import override_settings

from cms.api import add_plugin
from cms.models import CMSPlugin, Placeholder, UserSettings
from cms.test_utils.project.placeholderapp.models import Example1
from cms.test_utils.testcases import CMSTestCase


class AppAdminTestCase(CMSTestCase):

    def setUp(self):
        self._obj = self._get_example_obj()

    def _add_plugin_to_placeholder(self, placeholder,
                                   plugin_type='LinkPlugin', language='en'):
        plugin_data = {
            'StylePlugin': {'tag_type': 'div'},
            'LinkPlugin': {'name': 'A Link', 'external_link': 'https://www.django-cms.org'},
            'PlaceholderPlugin': {'name': 'Content'},
        }
        plugin = add_plugin(
            placeholder,
            plugin_type,
            language,
            **plugin_data[plugin_type]
        )
        return plugin

    def _get_move_data(self, plugin, position, placeholder=None, parent=None):
        try:
            placeholder_id = placeholder.pk
        except AttributeError:
            placeholder_id = ''

        try:
            parent_id = parent.pk
        except AttributeError:
            parent_id = ''

        data = {
            'placeholder_id': placeholder_id,
            'target_language': 'en',
            'target_position': position,
            'plugin_id': plugin.pk,
            'plugin_parent': parent_id,
        }
        return data

    def _get_add_plugin_uri(self, plugin_type='LinkPlugin', language='en'):
        uri = self.get_add_plugin_uri(
            placeholder=self._obj.placeholder,
            plugin_type=plugin_type,
            language=language,
        )
        return uri

    def _get_example_obj(self):
        obj = Example1.objects.create(
            char_1='one',
            char_2='two',
            char_3='tree',
            char_4='four'
        )
        return obj


class AppAdminTest(AppAdminTestCase):
    placeholderconf = {'placeholder': {
        'limits': {
            'global': 2,
            'StylePlugin': 1,
        }
    }
    }

    def test_global_limit_on_plugin_add(self):
        """
        Ensures placeholder global plugin limit is respected
        when adding plugins to the placeholder.
        """
        superuser = self.get_superuser()
        endpoint = self._get_add_plugin_uri()

        with self.login_user_context(superuser):
            with override_settings(CMS_PLACEHOLDER_CONF=self.placeholderconf):
                data = {'name': 'A Link', 'external_link': 'https://www.django-cms.org'}
                response = self.client.post(endpoint, data)  # first
                self.assertEqual(response.status_code, 200)
                response = self.client.post(endpoint, data)  # second
                self.assertEqual(response.status_code, 200)
                response = self.client.post(endpoint, data)  # third
                self.assertEqual(response.status_code, 400)
                self.assertEqual(
                    response.content,
                    b"This placeholder already has the maximum number of plugins (2).",
                )

    def test_global_limit_on_plugin_move(self):
        """
        Ensures placeholder global plugin limit is respected
        when moving plugins to the placeholder.
        """
        superuser = self.get_superuser()
        source_placeholder = self._obj.placeholder
        target_placeholder = self._get_example_obj().placeholder

        plugin_1 = self._add_plugin_to_placeholder(source_placeholder)
        plugin_2 = self._add_plugin_to_placeholder(source_placeholder)
        plugin_3 = self._add_plugin_to_placeholder(source_placeholder)

        with self.login_user_context(superuser):
            with self.settings(CMS_PLACEHOLDER_CONF=self.placeholderconf):
                data = self._get_move_data(plugin_1, position=1, placeholder=target_placeholder)
                endpoint = self.get_move_plugin_uri(plugin_1)
                response = self.client.post(endpoint, data)  # first
                self.assertEqual(response.status_code, 200)
                data = self._get_move_data(plugin_2, position=2, placeholder=target_placeholder)
                endpoint = self.get_move_plugin_uri(plugin_2)
                response = self.client.post(endpoint, data)  # second
                self.assertEqual(response.status_code, 200)
                data = self._get_move_data(plugin_3, position=3, placeholder=target_placeholder)
                endpoint = self.get_move_plugin_uri(plugin_3)
                response = self.client.post(endpoint, data)  # third
                self.assertEqual(response.status_code, 400)
                self.assertEqual(
                    response.content,
                    b"This placeholder already has the maximum number of plugins (2)."
                )

    def test_no_global_limit_check_same_placeholder_move(self):
        """
        Ensures no global limit exception is raised
        when moving plugins inside of a placeholder.
        """
        superuser = self.get_superuser()
        source_placeholder = self._obj.placeholder

        plugin_1 = self._add_plugin_to_placeholder(source_placeholder)
        plugin_2 = self._add_plugin_to_placeholder(source_placeholder)

        with self.login_user_context(superuser):
            with self.settings(CMS_PLACEHOLDER_CONF=self.placeholderconf):
                data = self._get_move_data(plugin_1, position=1)
                endpoint = self.get_move_plugin_uri(plugin_1)
                response = self.client.post(endpoint, data)  # first
                self.assertEqual(response.status_code, 200)
                data = self._get_move_data(plugin_2, position=2)
                endpoint = self.get_move_plugin_uri(plugin_2)
                response = self.client.post(endpoint, data)  # second
                self.assertEqual(response.status_code, 200)

    def test_type_limit_on_plugin_add(self):
        """
        Ensures placeholder plugin type limit is respected
        when adding plugins to the placeholder.
        """
        superuser = self.get_superuser()
        endpoint = self._get_add_plugin_uri('StylePlugin')

        with self.login_user_context(superuser):
            with self.settings(CMS_PLACEHOLDER_CONF=self.placeholderconf):
                data = {'tag_type': 'div'}
                response = self.client.post(endpoint, data)  # first
                self.assertEqual(response.status_code, 200)
                response = self.client.post(endpoint, data)  # second
                self.assertEqual(response.status_code, 400)
                self.assertEqual(
                    response.content,
                    b"This placeholder already has the "
                    b"maximum number (1) of allowed Style plugins."
                )

    def test_type_limit_on_plugin_move(self):
        """
        Ensures placeholder plugin type limit is respected
        when moving plugins to the placeholder.
        """
        superuser = self.get_superuser()
        source_placeholder = self._obj.placeholder
        target_placeholder = self._get_example_obj().placeholder

        plugin_1 = self._add_plugin_to_placeholder(source_placeholder, 'StylePlugin')
        plugin_2 = self._add_plugin_to_placeholder(source_placeholder, 'StylePlugin')

        with self.login_user_context(superuser):
            with self.settings(CMS_PLACEHOLDER_CONF=self.placeholderconf):
                data = self._get_move_data(plugin_1, position=1, placeholder=target_placeholder)
                endpoint = self.get_move_plugin_uri(plugin_1)
                response = self.client.post(endpoint, data)  # first
                self.assertEqual(response.status_code, 200)
                data = self._get_move_data(plugin_2, position=1, placeholder=target_placeholder)
                endpoint = self.get_move_plugin_uri(plugin_2)
                response = self.client.post(endpoint, data)  # second
                self.assertEqual(response.status_code, 400)
                self.assertEqual(response.content,
                                 b"This placeholder already has the maximum number (1) of allowed Style plugins.")

    def test_no_type_limit_check_same_placeholder_move(self):
        """
        Ensures no plugin type limit exception is raised
        when moving plugins inside of a placeholder.
        """
        superuser = self.get_superuser()
        source_placeholder = self._obj.placeholder

        plugin_1 = self._add_plugin_to_placeholder(source_placeholder, 'StylePlugin')

        with self.login_user_context(superuser):
            with self.settings(CMS_PLACEHOLDER_CONF=self.placeholderconf):
                data = self._get_move_data(plugin_1, position=1)
                endpoint = self.get_move_plugin_uri(plugin_1)
                response = self.client.post(endpoint, data)  # first
                self.assertEqual(response.status_code, 200)


class AppAdminPermissionsTest(AppAdminTestCase):

    def setUp(self):
        self._obj = self._get_example_obj()
        self._staff_user = self.get_staff_user_with_no_permissions()

    def test_user_can_add_plugin(self):
        """
        User can add a new plugin if he has change permissions
        on the model attached to the placeholder and he has
        add permissions on the plugin model.
        """
        staff_user = self._staff_user
        placeholder = self._obj.placeholder
        plugins = placeholder.get_plugins('en').filter(plugin_type='LinkPlugin')
        endpoint = self._get_add_plugin_uri()

        self.add_permission(staff_user, 'change_example1')
        self.add_permission(staff_user, 'add_link')

        with self.login_user_context(staff_user):
            data = {'name': 'A Link', 'external_link': 'https://www.django-cms.org'}
            response = self.client.post(endpoint, data)
            self.assertEqual(response.status_code, 200)
            self.assertEqual(plugins.count(), 1)

    def test_user_cant_add_plugin(self):
        """
        User can't add a new plugin if he does not have
        change permissions on the model attached to the placeholder
        and/or does not have add permissions on the plugin model.
        """
        staff_user = self._staff_user
        placeholder = self._obj.placeholder
        plugins = placeholder.get_plugins('en').filter(plugin_type='LinkPlugin')
        endpoint = self._get_add_plugin_uri()

        self.add_permission(staff_user, 'add_example1')
        self.add_permission(staff_user, 'delete_example1')
        self.add_permission(staff_user, 'add_link')

        with self.login_user_context(staff_user):
            data = {'name': 'A Link', 'external_link': 'https://www.django-cms.org'}
            response = self.client.post(endpoint, data)
            self.assertEqual(response.status_code, 403)
            self.assertEqual(plugins.count(), 0)

        self.add_permission(staff_user, 'change_example1')
        self.remove_permission(staff_user, 'add_link')

        with self.login_user_context(staff_user):
            data = {'name': 'A Link', 'external_link': 'https://www.django-cms.org'}
            response = self.client.post(endpoint, data)
            self.assertEqual(response.status_code, 403)
            self.assertEqual(plugins.count(), 0)

    def test_user_can_edit_plugin(self):
        """
        User can edit a plugin if he has change permissions
        on the model attached to the placeholder and he has
        change permissions on the plugin model.
        """
        staff_user = self._staff_user
        placeholder = self._obj.placeholder
        plugin = self._add_plugin_to_placeholder(placeholder)
        endpoint = self.get_change_plugin_uri(plugin)

        self.add_permission(staff_user, 'change_example1')
        self.add_permission(staff_user, 'change_link')

        with self.login_user_context(staff_user):
            data = model_to_dict(plugin, fields=['name', 'external_link'])
            data['name'] = 'A link 2'

            response = self.client.post(endpoint, data)
            self.assertEqual(response.status_code, 200)
            plugin.refresh_from_db()
            self.assertEqual(plugin.name, data['name'])

    def test_user_cant_edit_plugin(self):
        """
        User can't edit a plugin if he does not have
        change permissions on the model attached to the placeholder
        and/or does not have change permissions on the plugin model.
        """
        staff_user = self._staff_user
        placeholder = self._obj.placeholder
        plugin = self._add_plugin_to_placeholder(placeholder)
        endpoint = self.get_change_plugin_uri(plugin)

        self.add_permission(staff_user, 'add_example1')
        self.add_permission(staff_user, 'delete_example1')
        self.add_permission(staff_user, 'change_link')

        with self.login_user_context(staff_user):
            data = model_to_dict(plugin, fields=['name', 'external_link'])
            data['name'] = 'A link 2'

            response = self.client.post(endpoint, data)
            self.assertEqual(response.status_code, 403)
            plugin.refresh_from_db()
            self.assertNotEqual(plugin.name, data['name'])

        self.add_permission(staff_user, 'change_example1')
        self.remove_permission(staff_user, 'change_link')

        with self.login_user_context(staff_user):
            data = model_to_dict(plugin, fields=['name', 'external_link'])
            data['name'] = 'A link 2'

            response = self.client.post(endpoint, data)
            self.assertEqual(response.status_code, 403)
            plugin.refresh_from_db()
            self.assertNotEqual(plugin.name, data['name'])

    def test_user_can_delete_plugin(self):
        """
        User can delete a plugin if he has change permissions
        on the model attached to the placeholder and he has
        delete permissions on the plugin model.
        """
        staff_user = self._staff_user
        placeholder = self._obj.placeholder
        plugin = self._add_plugin_to_placeholder(placeholder)
        endpoint = self.get_delete_plugin_uri(plugin)

        self.add_permission(staff_user, 'change_example1')
        self.add_permission(staff_user, 'delete_link')

        with self.login_user_context(staff_user):
            data = {'post': True}

            response = self.client.post(endpoint, data)
            self.assertEqual(response.status_code, 302)
            self.assertFalse(CMSPlugin.objects.filter(pk=plugin.pk).exists())

    def test_user_cant_delete_plugin(self):
        """
        User can't delete a plugin if he does not have
        change permissions on the model attached to the placeholder
        and/or does not have delete permissions on the plugin model.
        """
        staff_user = self._staff_user
        placeholder = self._obj.placeholder
        plugin = self._add_plugin_to_placeholder(placeholder)
        endpoint = self.get_delete_plugin_uri(plugin)

        self.add_permission(staff_user, 'add_example1')
        self.add_permission(staff_user, 'delete_example1')
        self.add_permission(staff_user, 'delete_link')

        with self.login_user_context(staff_user):
            data = {'post': True}

            response = self.client.post(endpoint, data)
            self.assertEqual(response.status_code, 403)
            self.assertTrue(CMSPlugin.objects.filter(pk=plugin.pk).exists())

        self.add_permission(staff_user, 'change_example1')
        self.remove_permission(staff_user, 'delete_link')

        with self.login_user_context(staff_user):
            data = {'post': True}

            response = self.client.post(endpoint, data)
            self.assertEqual(response.status_code, 403)
            self.assertTrue(CMSPlugin.objects.filter(pk=plugin.pk).exists())

    def test_user_can_move_plugin(self):
        """
        User can move a plugin if he has change permissions
        on the model attached to the placeholder and he has
        change permissions on the plugin model.
        """
        staff_user = self._staff_user
        source_placeholder = self._obj.placeholder
        target_placeholder = self._get_example_obj().placeholder
        plugin = self._add_plugin_to_placeholder(source_placeholder)

        data = self._get_move_data(plugin, position=1, placeholder=target_placeholder)

        self.add_permission(staff_user, 'change_example1')
        self.add_permission(staff_user, 'change_link')

        with self.login_user_context(staff_user):
            endpoint = self.get_move_plugin_uri(plugin)
            response = self.client.post(endpoint, data)
            self.assertEqual(response.status_code, 200)
            self.assertTrue(target_placeholder.get_plugins('en').filter(pk=plugin.pk))
            self.assertFalse(source_placeholder.get_plugins('en').filter(pk=plugin.pk))

    def test_user_cant_move_plugin(self):
        """
        User can't move a plugin if he does not have
        change permissions on the model attached to the placeholder
        and/or does not have change permissions on the plugin model.
        """
        staff_user = self._staff_user
        source_placeholder = self._obj.placeholder
        target_placeholder = self._get_example_obj().placeholder
        plugin = self._add_plugin_to_placeholder(source_placeholder)

        data = self._get_move_data(plugin, position=1, placeholder=target_placeholder)

        self.add_permission(staff_user, 'add_example1')
        self.add_permission(staff_user, 'delete_example1')
        self.add_permission(staff_user, 'change_link')

        with self.login_user_context(staff_user):
            endpoint = self.get_move_plugin_uri(plugin)
            response = self.client.post(endpoint, data)
            self.assertEqual(response.status_code, 403)
            self.assertFalse(target_placeholder.get_plugins('en').filter(pk=plugin.pk))
            self.assertTrue(source_placeholder.get_plugins('en').filter(pk=plugin.pk))

        self.add_permission(staff_user, 'change_example1')
        self.remove_permission(staff_user, 'change_link')

        with self.login_user_context(staff_user):
            response = self.client.post(endpoint, data)
            self.assertEqual(response.status_code, 403)
            self.assertFalse(target_placeholder.get_plugins('en').filter(pk=plugin.pk))
            self.assertTrue(source_placeholder.get_plugins('en').filter(pk=plugin.pk))

    def test_user_can_copy_plugin(self):
        """
        User can copy a plugin if he has change permissions
        on the model attached to the placeholder and he has
        add permissions on the plugin model.
        """
        staff_user = self._staff_user
        placeholder = self._obj.placeholder
        plugin = self._add_plugin_to_placeholder(placeholder)
        endpoint = self.get_copy_plugin_uri(plugin)
        source_placeholder = plugin.placeholder
        target_placeholder = self._get_example_obj().placeholder

        data = {
            'source_plugin_id': plugin.pk,
            'source_placeholder_id': source_placeholder.pk,
            'source_language': plugin.language,
            'target_language': 'en',
            'target_placeholder_id': target_placeholder.pk,
        }

        self.add_permission(staff_user, 'change_example1')
        self.add_permission(staff_user, 'add_link')

        with self.login_user_context(staff_user):
            response = self.client.post(endpoint, data)
            self.assertEqual(response.status_code, 200)
            self.assertTrue(source_placeholder.get_plugins('en').filter(pk=plugin.pk).exists())
            self.assertTrue(
                target_placeholder
                .get_plugins('en')
                .filter(plugin_type=plugin.plugin_type)
                .exists()
            )

    def test_user_cant_copy_plugin(self):
        """
        User can't copy a plugin if he does not have
        change permissions on the model attached to the placeholder
        and/or does not have add permissions on the plugin model.
        """
        staff_user = self._staff_user
        placeholder = self._obj.placeholder
        plugin = self._add_plugin_to_placeholder(placeholder)
        endpoint = self.get_copy_plugin_uri(plugin)
        source_placeholder = plugin.placeholder
        target_placeholder = self._get_example_obj().placeholder

        data = {
            'source_plugin_id': plugin.pk,
            'source_placeholder_id': source_placeholder.pk,
            'source_language': plugin.language,
            'target_language': 'en',
            'target_placeholder_id': target_placeholder.pk,
        }

        self.add_permission(staff_user, 'add_example1')
        self.add_permission(staff_user, 'delete_example1')
        self.add_permission(staff_user, 'add_link')

        with self.login_user_context(staff_user):
            response = self.client.post(endpoint, data)
            self.assertEqual(response.status_code, 403)
            self.assertTrue(source_placeholder.get_plugins('en').filter(pk=plugin.pk).exists())
            self.assertFalse(
                target_placeholder
                .get_plugins('en')
                .filter(plugin_type=plugin.plugin_type)
                .exists()
            )

        self.add_permission(staff_user, 'change_example1')
        self.remove_permission(staff_user, 'add_link')

        with self.login_user_context(staff_user):
            response = self.client.post(endpoint, data)
            self.assertEqual(response.status_code, 403)
            self.assertTrue(source_placeholder.get_plugins('en').filter(pk=plugin.pk).exists())
            self.assertFalse(
                target_placeholder
                .get_plugins('en')
                .filter(plugin_type=plugin.plugin_type)
                .exists()
            )

    def test_user_can_clear_empty_placeholder(self):
        """
        User can clear a placeholder if he has change permissions
        on the model attached to the placeholder.
        """
        staff_user = self._staff_user
        placeholder = self._obj.placeholder
        endpoint = self.get_clear_placeholder_url(placeholder)

        self.add_permission(staff_user, 'change_example1')

        with self.login_user_context(staff_user):
            response = self.client.post(endpoint, {'test': 0})
            self.assertEqual(response.status_code, 302)

    def test_user_cant_clear_empty_placeholder(self):
        """
        User can't clear a placeholder if he does not have
        change permissions on the model attached to the placeholder.
        """
        staff_user = self._staff_user
        placeholder = self._obj.placeholder
        endpoint = self.get_clear_placeholder_url(placeholder)

        with self.login_user_context(staff_user):
            response = self.client.post(endpoint, {'test': 0})
            self.assertEqual(response.status_code, 403)

    def test_user_can_clear_non_empty_placeholder(self):
        """
        User can clear a placeholder with plugins if he has
        change permissions on the model attached to the placeholder
        and delete permissions on the plugin model.
        """
        staff_user = self._staff_user
        placeholder = self._obj.placeholder
        plugins = [
            self._add_plugin_to_placeholder(placeholder, 'StylePlugin'),
            self._add_plugin_to_placeholder(placeholder, 'LinkPlugin'),
        ]
        placeholder = plugins[0].placeholder
        endpoint = self.get_clear_placeholder_url(placeholder)

        self.add_permission(staff_user, 'delete_style')
        self.add_permission(staff_user, 'delete_link')
        self.add_permission(staff_user, 'change_example1')

        with self.login_user_context(staff_user):
            response = self.client.post(endpoint, {'test': 0})
            self.assertEqual(response.status_code, 302)
            self.assertEqual(placeholder.get_plugins('en').count(), 0)

    def test_user_cant_clear_non_empty_placeholder(self):
        """
        User can't clear a placeholder with plugins if he does not have
        change permissions on the model attached to the placeholder
        and/or does not have delete permissions on the plugin model.
        """
        staff_user = self._staff_user
        placeholder = self._obj.placeholder
        plugins = [
            self._add_plugin_to_placeholder(placeholder, 'StylePlugin'),
            self._add_plugin_to_placeholder(placeholder, 'LinkPlugin'),
        ]
        placeholder = plugins[0].placeholder
        endpoint = self.get_clear_placeholder_url(placeholder)

        self.add_permission(staff_user, 'delete_text')
        self.add_permission(staff_user, 'delete_link')

        with self.login_user_context(staff_user):
            response = self.client.post(endpoint, {'test': 0})
            self.assertEqual(response.status_code, 403)
            self.assertEqual(placeholder.get_plugins('en').count(), 2)

    def test_user_can_copy_placeholder_to_clipboard(self):
        """
        User can copy a placeholder to the clipboard
        if he has add permissions on the plugin models
        being copied.
        """
        staff_user = self._staff_user
        source_placeholder = self._obj.placeholder
        endpoint = self.get_copy_placeholder_uri(source_placeholder)

        self._add_plugin_to_placeholder(source_placeholder, 'StylePlugin')
        self._add_plugin_to_placeholder(source_placeholder, 'LinkPlugin')

        user_settings = UserSettings.objects.create(
            language="en",
            user=staff_user,
            clipboard=Placeholder.objects.create(),
        )

        self.add_permission(staff_user, 'add_link')
        self.add_permission(staff_user, 'add_style')

        data = {
            'source_plugin_id': '',
            'source_placeholder_id': source_placeholder.pk,
            'source_language': 'en',
            'target_language': 'en',
            'target_placeholder_id': user_settings.clipboard.pk,
        }

        with self.login_user_context(staff_user):
            # Copy plugins into the clipboard
            response = self.client.post(endpoint, data)
            self.assertEqual(response.status_code, 200)

        clipboard_plugins = user_settings.clipboard.get_plugins()

        # assert the clipboard has a PlaceholderPlugin
        self.assertTrue(clipboard_plugins.filter(plugin_type='PlaceholderPlugin').exists())

        self.assertEqual(len(clipboard_plugins), 1)

        placeholder_plugin = clipboard_plugins[0].get_plugin_instance()[0]
        ref_placeholder = placeholder_plugin.placeholder_ref

        # assert there's only two plugins in the clipboard
        self.assertEqual(ref_placeholder.get_plugins().count(), 2)

    def test_user_cant_copy_placeholder_to_clipboard(self):
        """
        User cant copy a placeholder to the clipboard if he does not
        have add permissions on the plugin models being copied.
        """
        staff_user = self._staff_user
        source_placeholder = self._obj.placeholder
        endpoint = self.get_copy_placeholder_uri(source_placeholder)

        self._add_plugin_to_placeholder(source_placeholder, 'StylePlugin')
        self._add_plugin_to_placeholder(source_placeholder, 'LinkPlugin')

        user_settings = UserSettings.objects.create(
            language="en",
            user=staff_user,
            clipboard=Placeholder.objects.create(),
        )

        self.add_permission(staff_user, 'change_link')
        self.add_permission(staff_user, 'delete_link')
        self.add_permission(staff_user, 'change_style')
        self.add_permission(staff_user, 'delete_style')

        data = {
            'source_plugin_id': '',
            'source_placeholder_id': source_placeholder.pk,
            'source_language': 'en',
            'target_language': 'en',
            'target_placeholder_id': user_settings.clipboard.pk,
        }

        with self.login_user_context(staff_user):
            # Copy plugins into the clipboard
            response = self.client.post(endpoint, data)

        self.assertEqual(response.status_code, 403)

        clipboard_plugins = user_settings.clipboard.get_plugins()

        self.assertEqual(len(clipboard_plugins), 0)

    def test_user_can_paste_from_clipboard(self):
        """
        User can paste plugins from the clipboard if he has
        change permissions on the model attached to the target
        placeholder and he has add permissions on the plugin models
        being copied.
        """
        staff_user = self._staff_user
        target_placeholder = self._obj.placeholder

        self.add_permission(staff_user, 'change_example1')
        self.add_permission(staff_user, 'add_link')

        user_settings = UserSettings.objects.create(
            language="en",
            user=staff_user,
            clipboard=Placeholder.objects.create(),
        )

        placeholder_plugin = self._add_plugin_to_placeholder(
            user_settings.clipboard,
            'PlaceholderPlugin',
        )
        ref_placeholder = placeholder_plugin.placeholder_ref

        self._add_plugin_to_placeholder(ref_placeholder)
        self._add_plugin_to_placeholder(ref_placeholder)

        with self.login_user_context(staff_user):
            # Paste plugins from clipboard into placeholder
            # under the french language.
            data = {
                'placeholder_id': target_placeholder.pk,
                'plugin_id': placeholder_plugin.pk,
                'plugin_parent': '',
                'target_language': 'fr',
                'move_a_copy': True,
                'target_position': target_placeholder.get_next_plugin_position('fr', insert_order='last'),
            }
            endpoint = self.get_move_plugin_uri(placeholder_plugin)
            response = self.client.post(endpoint, data)
            self.assertEqual(response.status_code, 200)
            self.assertEqual(target_placeholder.get_plugins('fr').count(), 2)

    def test_user_cant_paste_from_clipboard(self):
        """
        User cant paste plugins from the clipboard if he does not have
        change permissions on the model attached to the target placeholder
        and/or does not have add permissions on the plugin models
        being copied.
        """
        staff_user = self._staff_user
        target_placeholder = self._obj.placeholder

        self.add_permission(staff_user, 'add_example1')
        self.add_permission(staff_user, 'delete_example1')
        self.add_permission(staff_user, 'add_link')

        user_settings = UserSettings.objects.create(
            language="en",
            user=staff_user,
            clipboard=Placeholder.objects.create(),
        )

        placeholder_plugin = self._add_plugin_to_placeholder(
            user_settings.clipboard,
            'PlaceholderPlugin',
        )
        ref_placeholder = placeholder_plugin.placeholder_ref

        self._add_plugin_to_placeholder(ref_placeholder)
        self._add_plugin_to_placeholder(ref_placeholder)

        with self.login_user_context(staff_user):
            # Paste plugins from clipboard into placeholder
            # under the french language.
            data = {
                'placeholder_id': target_placeholder.pk,
                'plugin_id': placeholder_plugin.pk,
                'plugin_parent': '',
                'target_language': 'fr',
                'move_a_copy': True,
                'target_position': target_placeholder.get_next_plugin_position('fr', insert_order='last'),
            }
            endpoint = self.get_move_plugin_uri(placeholder_plugin)
            response = self.client.post(endpoint, data)
            self.assertEqual(response.status_code, 403)
            self.assertEqual(target_placeholder.get_plugins('fr').count(), 0)

        self.add_permission(staff_user, 'change_example1')
        self.remove_permission(staff_user, 'add_link')

        with self.login_user_context(staff_user):
            # Paste plugins from clipboard into placeholder
            # under the french language.
            data = {
                'placeholder_id': target_placeholder.pk,
                'plugin_id': placeholder_plugin.pk,
                'plugin_parent': '',
                'target_language': 'fr',
                'move_a_copy': True,
                'target_position': target_placeholder.get_next_plugin_position('fr', insert_order='last'),
            }
            response = self.client.post(endpoint, data)
            self.assertEqual(response.status_code, 403)
            self.assertEqual(target_placeholder.get_plugins('fr').count(), 0)



from django import forms
from django.contrib.auth.forms import AuthenticationForm


class CMSToolbarLoginForm(AuthenticationForm):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        kwargs['prefix'] = kwargs.get('prefix', 'cms')
        self.fields['username'].widget = forms.TextInput(
            attrs={'required': 'required'}
        )
        self.fields['password'].widget = forms.PasswordInput(
            attrs={'required': 'required'}
        )



from django.core.management.base import CommandError

from cms.utils.check import FileOutputWrapper, check

from .base import SubcommandsCommand


class CheckInstallation(SubcommandsCommand):
    help_string = 'Checks your settings and environment'
    command_name = 'check'

    def handle(self, *args, **options):
        if not check(FileOutputWrapper(self.stdout, self.stderr)):
            raise CommandError()



from cms.app_base import CMSAppConfig, CMSAppExtension


class CMSFeatureConfig(CMSAppExtension):
    pass


class CMSFeatureConfig2(CMSAppExtension):
    pass


class CMSConfig(CMSAppConfig):
    pass



from django.conf import settings
from django.http import Http404
from django.template.response import TemplateResponse
from django.urls import Resolver404, resolve, reverse

from cms import __version__, constants
from cms.cache.page import set_page_cache
from cms.models import EmptyPageContent
from cms.utils.page import get_page_template_from_request
from cms.utils.page_permissions import user_can_change_page, user_can_view_page


def render_page(request, page, current_language, slug):
    """
    Renders a page
    """
    context = {}
    context['lang'] = current_language
    context['current_page'] = page
    context['has_change_permissions'] = user_can_change_page(request.user, page)
    context['has_view_permissions'] = user_can_view_page(request.user, page)

    cant_view_page = any([
        not context['has_view_permissions'],
        isinstance(page.get_content_obj(current_language), EmptyPageContent)
    ])
    if cant_view_page:
        return _handle_no_page(request)

    template = get_page_template_from_request(request)
    response = TemplateResponse(request, template, context)
    response.add_post_render_callback(set_page_cache)

    # Add headers for X Frame Options - this really should be changed upon moving to class based views
    xframe_options = page.get_xframe_options()
    # xframe_options can be None if there's no xframe information on the page
    # (eg. a top-level page which has xframe options set to "inherit")
    if xframe_options == constants.X_FRAME_OPTIONS_INHERIT or xframe_options is None:
        # This is when we defer to django's own clickjacking handling
        return response

    # We want to prevent django setting this in their middlewear
    response.xframe_options_exempt = True

    if xframe_options == constants.X_FRAME_OPTIONS_ALLOW:
        # Do nothing, allowed is no header.
        return response
    elif xframe_options == constants.X_FRAME_OPTIONS_SAMEORIGIN:
        response['X-Frame-Options'] = 'SAMEORIGIN'
    elif xframe_options == constants.X_FRAME_OPTIONS_DENY:
        response['X-Frame-Options'] = 'DENY'
    return response


def _handle_no_page(request):
    try:
        # add a $ to the end of the url (does not match on the cms anymore)
        resolve('%s$' % request.path)
    except Resolver404 as e:
        # raise a django http 404 page
        exc = Http404(dict(path=request.path, tried=e.args[0]['tried']))
        raise exc
    raise Http404('CMS Page not found: %s' % request.path)


def _handle_no_apphook(request):
    context = {
        "absolute_url": request.toolbar.request_path,
    }
    return TemplateResponse(request, "cms/noapphook.html", context)


def _render_welcome_page(request):
    context = {
        'cms_version': __version__,
        'django_debug': settings.DEBUG,
        'next_url': reverse('pages-root'),
    }
    return TemplateResponse(request, "cms/welcome.html", context)


def render_pagecontent(request, pagecontent):
    language = pagecontent.language
    request.current_page = page = pagecontent.page
    page.page_content_cache[language] = pagecontent
    return render_page(request, page, language, page.get_slug(language))



import json

from django.http import HttpResponse
from django.utils.encoding import smart_str

NOT_FOUND_RESPONSE = "NotFound"


def jsonify_request(response):
    """ Turn any response in a 200 response to let jQuery code handle it nicely.
        Response contains a json object with the following attributes:
         * status: original response status code
         * content: original response content
    """
    content = {'status': response.status_code, 'content': smart_str(response.content, response.charset)}
    return HttpResponse(json.dumps(content), content_type="application/json")



# TODO: this is just stuff from utils.py - should be splitted / moved
from django.conf import settings
from django.core.files.storage import get_storage_class
from django.utils.functional import LazyObject

from cms.utils.i18n import (
    get_default_language,
    get_language_code,
    get_language_list,
)


def get_current_site():
    from django.contrib.sites.models import Site

    return Site.objects.get_current()


def get_language_from_request(request, current_page=None):
    """
    Return the most obvious language according the request
    """
    language = None
    if hasattr(request, 'POST'):
        language = request.POST.get('language', None)
    if hasattr(request, 'GET') and not language:
        language = request.GET.get('language', None)
    site_id = current_page.node.site_id if current_page else None
    if language:
        language = get_language_code(language)
        if language not in get_language_list(site_id):
            language = None
    if not language:
        language = get_language_code(getattr(request, 'LANGUAGE_CODE', None))
    if language:
        if language not in get_language_list(site_id):
            language = None

    if not language and current_page:
        # in last resort, get the first language available in the page
        languages = current_page.get_languages()

        if len(languages) > 0:
            language = languages[0]

    if not language:
        # language must be defined in CMS_LANGUAGES, so check first if there
        # is any language with LANGUAGE_CODE, otherwise try to split it and find
        # best match
        language = get_default_language(site_id=site_id)

    return language


default_storage = 'django.contrib.staticfiles.storage.StaticFilesStorage'


class ConfiguredStorage(LazyObject):
    def _setup(self):
        self._wrapped = get_storage_class(getattr(settings, 'STATICFILES_STORAGE', default_storage))()


configured_storage = ConfiguredStorage()



from django.conf import settings

from cms.utils.conf import get_cms_setting


def _site_cache_key(lang):
    return "%s-%s" % (get_cms_setting('SITE_CHOICES_CACHE_KEY'), lang)


def _page_cache_key(lang):
    return "%s-%s" % (get_cms_setting('PAGE_CHOICES_CACHE_KEY'), lang)


def _clean_many(prefix):
    from django.core.cache import cache
    keys = []
    if settings.USE_I18N:
        for lang in [language[0] for language in settings.LANGUAGES]:
            keys.append("%s-%s" % (prefix, lang))
    else:
        keys = ["%s-%s" % (prefix, settings.LANGUAGE_CODE)]
    cache.delete_many(keys)


def clean_site_choices_cache(sender, **kwargs):
    _clean_many(get_cms_setting('SITE_CHOICES_CACHE_KEY'))


def clean_page_choices_cache(sender, **kwargs):
    _clean_many(get_cms_setting('PAGE_CHOICES_CACHE_KEY'))



from django.db import migrations

try:
    IrreversibleError = migrations.Migration.IrreversibleError
except AttributeError:
    from django.db.migrations.exceptions import IrreversibleError


class IrreversibleMigration(migrations.Migration):

    def unapply(self, project_state, schema_editor, collect_sql=False):
        raise IrreversibleError('Migration %s is not reversible' % self.name)



import operator
import warnings
from collections import OrderedDict

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.db.models.query_utils import Q
from django.template import (
    Context,
    NodeList,
    Template,
    TemplateSyntaxError,
    Variable,
    engines,
)
from django.template.base import VariableNode
from django.template.loader import get_template
from django.template.loader_tags import BlockNode, ExtendsNode, IncludeNode
from sekizai.helpers import get_varname

from cms.exceptions import DuplicatePlaceholderWarning
from cms.models import Placeholder
from cms.utils.conf import get_cms_setting


def _get_nodelist(tpl):
    if hasattr(tpl, 'template'):
        return tpl.template.nodelist
    else:
        return tpl.nodelist


def get_context():
    if engines is not None:
        context = Context()
        context.template = Template('')
        return context
    else:
        return {}


def get_placeholder_conf(setting, placeholder, template=None, default=None):
    """
    Returns the placeholder configuration for a given setting. The key would for
    example be 'plugins' or 'name'.

    Resulting value will be the last from:

    CMS_PLACEHOLDER_CONF[None] (global)
    CMS_PLACEHOLDER_CONF['template'] (if template is given)
    CMS_PLACEHOLDER_CONF['placeholder']
    CMS_PLACEHOLDER_CONF['template placeholder'] (if template is given)
    """

    if placeholder:
        keys = []
        placeholder_conf = get_cms_setting('PLACEHOLDER_CONF')
        # 1st level
        if template:
            keys.append(u'%s %s' % (template, placeholder))
        # 2nd level
        keys.append(placeholder)
        # 3rd level
        if template:
            keys.append(template)
        # 4th level
        keys.append(None)
        for key in keys:
            try:
                conf = placeholder_conf[key]
                value = conf.get(setting, None)
                if value is not None:
                    return value
                inherit = conf.get('inherit')
                if inherit:
                    if ' ' in inherit:
                        inherit = inherit.split(' ')
                    else:
                        inherit = (None, inherit)
                    value = get_placeholder_conf(setting, inherit[1], inherit[0], default)
                    if value is not None:
                        return value
            except KeyError:
                continue
    return default


def get_toolbar_plugin_struct(plugins, slot=None, page=None):
    """
    Return the list of plugins to render in the toolbar.
    The dictionary contains the label, the classname and the module for the
    plugin.
    Names and modules can be defined on a per-placeholder basis using
    'plugin_modules' and 'plugin_labels' attributes in CMS_PLACEHOLDER_CONF

    :param plugins: list of plugins
    :param slot: placeholder slot name
    :param page: the page
    :return: list of dictionaries
    """
    template = None

    if page:
        template = page.get_template()

    modules = get_placeholder_conf("plugin_modules", slot, template, default={})
    names = get_placeholder_conf("plugin_labels", slot, template, default={})

    main_list = []

    # plugin.value points to the class name of the plugin
    # It's added on registration. TIL.
    for plugin in plugins:
        main_list.append({'value': plugin.value,
                          'name': names.get(plugin.value, plugin.name),
                          'module': modules.get(plugin.value, plugin.module)})
    return sorted(main_list, key=operator.itemgetter("module"))


def validate_placeholder_name(name):
    if not isinstance(name, str):
        raise ImproperlyConfigured("Placeholder identifier names need to be of type string. ")

    if not all(ord(char) < 128 for char in name):
        raise ImproperlyConfigured("Placeholder identifiers names may not "
                                   "contain non-ascii characters. If you wish your placeholder "
                                   "identifiers to contain non-ascii characters when displayed to "
                                   "users, please use the CMS_PLACEHOLDER_CONF setting with the 'name' "
                                   "key to specify a verbose name.")


class PlaceholderNoAction():
    can_copy = False

    def copy(self, **kwargs):
        return False

    def get_copy_languages(self, **kwargs):
        return []


class MLNGPlaceholderActions(PlaceholderNoAction):
    can_copy = True

    def copy(self, target_placeholder, source_language, fieldname, model, target_language, **kwargs):
        from cms.utils.plugins import copy_plugins_to_placeholder
        trgt = model.objects.get(**{fieldname: target_placeholder})
        src = model.objects.get(master=trgt.master, language_code=source_language)

        source_placeholder = getattr(src, fieldname, None)
        if not source_placeholder:
            return False
        return copy_plugins_to_placeholder(
            source_placeholder.get_plugins_list(),
            placeholder=target_placeholder,
            language=target_language
        )

    def get_copy_languages(self, placeholder, model, fieldname, **kwargs):
        manager = model.objects
        src = manager.get(**{fieldname: placeholder})
        query = Q(master=src.master)
        query &= Q(**{'%s__cmsplugin__isnull' % fieldname: False})
        query &= ~Q(pk=src.pk)

        language_codes = manager.filter(query).values_list('language_code', flat=True).distinct()
        return [(lc, dict(settings.LANGUAGES)[lc]) for lc in language_codes]


def restore_sekizai_context(context, changes):
    varname = get_varname()
    sekizai_container = context.get(varname)
    for key, values in changes.items():
        sekizai_namespace = sekizai_container[key]
        for value in values:
            sekizai_namespace.append(value)


def _scan_placeholders(nodelist, node_class=None, current_block=None, ignore_blocks=None):
    from cms.templatetags.cms_tags import Placeholder

    if not node_class:
        node_class = Placeholder

    nodes = []

    if ignore_blocks is None:
        # List of BlockNode instances to ignore.
        # This is important to avoid processing overriden block nodes.
        ignore_blocks = []

    for node in nodelist:
        # check if this is a placeholder first
        if isinstance(node, node_class):
            nodes.append(node)
        elif isinstance(node, IncludeNode):
            # if there's an error in the to-be-included template, node.template becomes None
            if node.template:
                # Check if it quacks like a template object, if not
                # presume is a template path and get the object out of it
                if not callable(getattr(node.template, 'render', None)):
                    # If it's a variable there is no way to expand it at this stage so we
                    # need to skip it
                    if isinstance(node.template.var, Variable):
                        continue
                    else:
                        template = get_template(node.template.var)
                else:
                    template = node.template
                nodes += _scan_placeholders(_get_nodelist(template), node_class, current_block)
        # handle {% extends ... %} tags
        elif isinstance(node, ExtendsNode):
            nodes += _get_placeholder_nodes_from_extend(node, node_class)
        # in block nodes we have to scan for super blocks
        elif isinstance(node, VariableNode) and current_block:
            if node.filter_expression.token == 'block.super':
                if not hasattr(current_block.super, 'nodelist'):
                    raise TemplateSyntaxError("Cannot render block.super for blocks without a parent.")
                nodes += _scan_placeholders(_get_nodelist(current_block.super), node_class, current_block.super)
        # ignore nested blocks which are already handled
        elif isinstance(node, BlockNode) and node.name in ignore_blocks:
            continue
        # if the node has the newly introduced 'child_nodelists' attribute, scan
        # those attributes for nodelists and recurse them
        elif hasattr(node, 'child_nodelists'):
            for nodelist_name in node.child_nodelists:
                if hasattr(node, nodelist_name):
                    subnodelist = getattr(node, nodelist_name)
                    if isinstance(subnodelist, NodeList):
                        if isinstance(node, BlockNode):
                            current_block = node
                        nodes += _scan_placeholders(subnodelist, node_class, current_block, ignore_blocks)
        # else just scan the node for nodelist instance attributes
        else:
            for attr in dir(node):
                obj = getattr(node, attr)
                if isinstance(obj, NodeList):
                    if isinstance(node, BlockNode):
                        current_block = node
                    nodes += _scan_placeholders(obj, node_class, current_block, ignore_blocks)
    return nodes


def _scan_static_placeholders(nodelist):
    from cms.templatetags.cms_tags import StaticPlaceholderNode

    return _scan_placeholders(nodelist, node_class=StaticPlaceholderNode)


def get_placeholders(template):
    compiled_template = get_template(template)

    placeholders = []
    nodes = _scan_placeholders(_get_nodelist(compiled_template))
    clean_placeholders = []

    for node in nodes:
        placeholder = node.get_declaration()
        slot = placeholder.slot

        if slot in clean_placeholders:
            warnings.warn("Duplicate {{% placeholder \"{0}\" %}} "
                          "in template {1}."
                          .format(slot, template),
                          DuplicatePlaceholderWarning)
        else:
            validate_placeholder_name(slot)
            placeholders.append(placeholder)
            clean_placeholders.append(slot)
    return placeholders


def get_static_placeholders(template, context):
    compiled_template = get_template(template)
    nodes = _scan_static_placeholders(_get_nodelist(compiled_template))
    placeholders = [node.get_declaration(context) for node in nodes]
    placeholders_with_code = []

    for placeholder in placeholders:
        if placeholder.slot:
            placeholders_with_code.append(placeholder)
        else:
            warnings.warn('Unable to resolve static placeholder '
                          'name in template "{}"'.format(template),
                          Warning)
    return placeholders_with_code


def _get_block_nodes(extend_node):
    parent = extend_node.get_parent(get_context())
    parent_nodelist = _get_nodelist(parent)
    parent_nodes = parent_nodelist.get_nodes_by_type(BlockNode)
    parent_extend_nodes = parent_nodelist.get_nodes_by_type(ExtendsNode)

    if parent_extend_nodes:
        # Start at the top
        # Scan the extends node from the parent (if any)
        nodes = _get_block_nodes(parent_extend_nodes[0])
    else:
        nodes = OrderedDict()

    # Continue with the parent template nodes
    for node in parent_nodes:
        nodes[node.name] = node

    # Move on to the current template nodes
    current_nodes = _get_nodelist(extend_node).get_nodes_by_type(BlockNode)

    for node in current_nodes:
        if node.name in nodes:
            # set this node as the super node (for {{ block.super }})
            node.super = nodes[node.name]
        nodes[node.name] = node
    return nodes


def _get_placeholder_nodes_from_extend(extend_node, node_class):
    """
    Returns a list of placeholders found in the parent template(s) of this
    ExtendsNode
    """
    # This is a dictionary mapping all BlockNode instances found
    # in the template that contains the {% extends %} tag
    block_nodes = _get_block_nodes(extend_node)
    block_names = list(block_nodes.keys())

    placeholders = []

    for block in block_nodes.values():
        placeholders.extend(_scan_placeholders(_get_nodelist(block), node_class, block, block_names))

    # Scan topmost template for placeholder outside of blocks
    parent_template = _find_topmost_template(extend_node)
    placeholders += _scan_placeholders(_get_nodelist(parent_template), node_class, None, block_names)
    return placeholders


def _find_topmost_template(extend_node):
    parent_template = extend_node.get_parent(get_context())
    for node in _get_nodelist(parent_template).get_nodes_by_type(ExtendsNode):
        # Their can only be one extend block in a template, otherwise django raises an exception
        return _find_topmost_template(node)
        # No ExtendsNode
    return extend_node.get_parent(get_context())


def rescan_placeholders_for_obj(obj):
    from cms.models import Placeholder

    existing = OrderedDict()
    declared_placeholders = get_declared_placeholders_for_obj(obj)
    placeholders = [pl.slot for pl in declared_placeholders]

    for placeholder in Placeholder.objects.get_for_obj(obj):
        if placeholder.slot in placeholders:
            existing[placeholder.slot] = placeholder

    for placeholder in placeholders:
        if placeholder not in existing:
            existing[placeholder] = Placeholder.objects.create(
                slot=placeholder,
                source=obj,
            )
    return existing


def get_declared_placeholders_for_obj(obj):
    """Returns declared placeholders for an object. The object is supposed to have a method ``get_template``
    which returns the template path as a string that renders the object. ``get_declared_placeholders`` returns
    a list of placeholders used in the template by the ``{% placeholder %}`` template tag."""
    if not hasattr(obj, 'get_template'):
        raise NotImplementedError('%s should implement get_template' % obj.__class__.__name__)
    return get_placeholders(obj.get_template())


def get_placeholder_from_slot(placeholder_relation: models.Manager, slot: str, template_obj=None) -> Placeholder:
    """Retrieves the placeholder instance for a PlaceholderRelationField either by scaning the template
    of the template_obj (if given) or by creating or getting a Placeholder in the database"""
    if hasattr(template_obj, "get_template"):
        # Tries to get a placeholder (based on the template for the template_obj
        # or - if non exists - raises a Placeholder.DoesNotExist exception
        # Placeholders are marked in the template with {% placeholder %}
        try:
            return placeholder_relation.get(slot=slot)
        except Placeholder.DoesNotExist:
            rescan_placeholders_for_obj(template_obj)
            return placeholder_relation.placeholders.get(slot=slot)
    else:
        # Gets or creates the placeholder in any model. Placeholder is
        # rendered by {% render_placeholder %}
        return placeholder_relation.get_or_create(slot=slot)[0]



from django.forms.models import model_to_dict
from mock import Mock, call, patch

from cms.api import create_page
from cms.models import Placeholder, UserSettings, fields
from cms.models.fields import PlaceholderRelationField
from cms.test_utils.testcases import CMSTestCase


class ChecksTestCase(CMSTestCase):

    def test_no_check_means_passed(self):
        superuser = self.get_superuser()
        field = PlaceholderRelationField()
        page = create_page('test', 'nav_playground.html', 'en')
        placeholder = page.get_placeholders('en').get(slot='body')
        mock = Mock(return_value=[])
        with patch.object(PlaceholderRelationField, 'checks', property(mock)):
            self.assertTrue(field.run_checks(placeholder, superuser))

    def test_checks(self):
        """Test that PlaceholderRelationField.checks combines
        default_checks with field-level checks.
        """
        check1 = Mock(return_value=True)
        check2 = Mock(return_value=True)
        check3 = Mock(return_value=True)
        field = PlaceholderRelationField(checks=[check2, check3])
        with patch.object(field, 'default_checks', [check1]):
            self.assertEqual(
                list(field.checks),
                [check1, check2, check3],
            )

    def test_checks_one_false(self):
        """Test that one False value short-circuits.
        """

        superuser = self.get_superuser()
        page = create_page('test', 'nav_playground.html', 'en')
        placeholder = page.get_placeholders('en').get(slot='body')

        check1 = Mock(return_value=False)
        check2 = Mock(return_value=True)
        field = PlaceholderRelationField(checks=[check1, check2])
        self.assertFalse(field.run_checks(placeholder, superuser))
        check1.assert_called_once_with(placeholder, superuser)
        check2.assert_not_called()


class ChecksPlaceholderInterfaceTestCase(CMSTestCase):

    def test_check_source(self):
        superuser = self.get_superuser()
        page = create_page('test', 'nav_playground.html', 'en')
        placeholder = page.get_placeholders('en').get(slot='body')
        check = Mock(return_value=True)
        mock = Mock(return_value=[check])
        with patch.object(PlaceholderRelationField, 'checks', property(mock)):
            self.assertTrue(placeholder.check_source(superuser))
        check.assert_called_once_with(placeholder, superuser)

    def test_check_source_when_source_is_empty(self):
        superuser = self.get_superuser()
        placeholder = Placeholder.objects.create(slot='test')
        check = Mock(return_value=True)
        mock = Mock(return_value=[check])
        with patch.object(PlaceholderRelationField, 'checks', property(mock)):
            self.assertTrue(placeholder.check_source(superuser))
        check.assert_not_called()


class ChecksUsedInAdminEndpointsTestCase(CMSTestCase):

    def setUp(self):
        self.check = Mock(return_value=True)
        self.original_checks = fields.PlaceholderRelationField.default_checks
        fields.PlaceholderRelationField.default_checks = self.original_checks + [self.check]

    def tearDown(self):
        fields.PlaceholderRelationField.default_checks = self.original_checks

    def _get_move_data(self, plugin, position, placeholder=None, parent=None):
        try:
            placeholder_id = placeholder.pk
        except AttributeError:
            placeholder_id = ''

        try:
            parent_id = parent.pk
        except AttributeError:
            parent_id = ''

        data = {
            'placeholder_id': placeholder_id,
            'target_language': 'en',
            'target_position': position,
            'plugin_id': plugin.pk,
            'plugin_parent': parent_id,
        }
        return data

    def test_add_plugin_to_placeholder_without_source(self):
        superuser = self.get_superuser()
        placeholder = Placeholder.objects.create(slot='test')
        uri = self.get_add_plugin_uri(
            placeholder=placeholder,
            plugin_type='LinkPlugin',
            language='en',
        )

        with self.login_user_context(superuser):
            data = {'name': 'A Link', 'external_link': 'https://www.django-cms.org'}
            self.client.post(uri, data)
        self.check.assert_not_called()

    def test_add_plugin(self):
        superuser = self.get_superuser()
        page = create_page('test', 'nav_playground.html', 'en')
        placeholder = page.get_placeholders('en').get(slot='body')
        uri = self.get_add_plugin_uri(
            placeholder=placeholder,
            plugin_type='LinkPlugin',
            language='en',
        )

        with self.login_user_context(superuser):
            data = {'name': 'A Link', 'external_link': 'https://www.django-cms.org'}
            self.client.post(uri, data)
        self.check.assert_called_once_with(placeholder, superuser)

    def test_copy_plugins_to_clipboard(self):
        superuser = self.get_superuser()
        source_placeholder = Placeholder.objects.create(slot='test')
        source_plugin = self._add_plugin_to_placeholder(source_placeholder)
        uri = self.get_copy_plugin_uri(source_plugin)
        user_settings = UserSettings.objects.create(
            language='en',
            user=superuser,
            clipboard=Placeholder.objects.create(),
        )

        with self.login_user_context(superuser):
            data = {
                'source_language': 'en',
                'source_placeholder_id': source_placeholder.pk,
                'source_plugin_id': source_plugin.pk,
                'target_language': 'en',
                'target_placeholder_id': user_settings.clipboard.pk,
            }
            self.client.post(uri, data)

        self.check.assert_not_called()

    def test_copy_placeholder_to_clipboard(self):
        superuser = self.get_superuser()
        source_placeholder = Placeholder.objects.create(slot='test')
        source_plugin = self._add_plugin_to_placeholder(source_placeholder)
        uri = self.get_copy_plugin_uri(source_plugin)
        user_settings = UserSettings.objects.create(
            language='en',
            user=superuser,
            clipboard=Placeholder.objects.create(),
        )

        with self.login_user_context(superuser):
            data = {
                'source_language': "en",
                'source_placeholder_id': source_placeholder.pk,
                'target_language': "en",
                'target_placeholder_id': user_settings.clipboard.pk,
            }
            self.client.post(uri, data)

        self.check.assert_not_called()

    def test_copy_plugins_to_placeholder_without_source(self):
        superuser = self.get_superuser()
        page = create_page('test', 'nav_playground.html', 'en')
        source_placeholder = page.get_placeholders('en').get(slot='body')
        source_plugin = self._add_plugin_to_placeholder(source_placeholder)
        target_placeholder = Placeholder.objects.create(slot='test')
        uri = self.get_copy_plugin_uri(source_plugin)

        with self.login_user_context(superuser):
            data = {
                'source_language': "en",
                'source_placeholder_id': source_placeholder.pk,
                'source_plugin_id': source_plugin.pk,
                'target_language': "en",
                'target_placeholder_id': target_placeholder.pk,
            }
            self.client.post(uri, data)
        self.check.assert_not_called()

    def test_edit_plugin(self):
        superuser = self.get_superuser()
        page = create_page('test', 'nav_playground.html', 'en')
        placeholder = page.get_placeholders('en').get(slot='body')
        plugin = self._add_plugin_to_placeholder(placeholder)
        uri = self.get_change_plugin_uri(plugin)

        with self.login_user_context(superuser):
            data = model_to_dict(plugin, fields=['name', 'external_link'])
            data['name'] = 'Contents modified'
            self.client.post(uri, data)

        self.check.assert_called_once_with(placeholder, superuser)

    def test_move_plugin(self):
        superuser = self.get_superuser()
        page = create_page('test', 'nav_playground.html', 'en')
        page2 = create_page('test', 'nav_playground.html', 'en')
        source_placeholder = page.get_placeholders('en').get(slot='body')
        target_placeholder = page2.get_placeholders('en').get(slot='body')
        source_plugin = self._add_plugin_to_placeholder(source_placeholder)
        uri = self.get_move_plugin_uri(source_plugin)

        with self.login_user_context(superuser):
            data = self._get_move_data(
                source_plugin,
                position=1,
                placeholder=target_placeholder,
            )
            self.client.post(uri, data)

        self.assertEqual(self.check.call_count, 2)
        self.check.assert_has_calls([
            call(source_placeholder, superuser),
            call(target_placeholder, superuser),
        ])

    def test_cut_plugin(self):
        superuser = self.get_superuser()
        page = create_page('test', 'nav_playground.html', 'en')
        source_placeholder = page.get_placeholders('en').get(slot='body')
        source_plugin = self._add_plugin_to_placeholder(source_placeholder)
        uri = self.get_move_plugin_uri(source_plugin)
        user_settings = UserSettings.objects.create(
            language='en',
            user=superuser,
            clipboard=Placeholder.objects.create(),
        )

        with self.login_user_context(superuser):
            data = self._get_move_data(
                source_plugin,
                position=1,
                placeholder=user_settings.clipboard,
            )
            self.client.post(uri, data)

        self.check.assert_called_once_with(source_placeholder, superuser)

    def test_delete_plugin(self):
        superuser = self.get_superuser()
        page = create_page('test', 'nav_playground.html', 'en')
        placeholder = page.get_placeholders('en').get(slot='body')
        plugin = self._add_plugin_to_placeholder(placeholder)
        uri = self.get_delete_plugin_uri(plugin)

        with self.login_user_context(superuser):
            data = {'post': True}
            self.client.post(uri, data)

        self.check.assert_called_once_with(placeholder, superuser)

    def test_clear_placeholder(self):
        superuser = self.get_superuser()
        page = create_page('test', 'nav_playground.html', 'en')
        placeholder = page.get_placeholders('en').get(slot='body')
        uri = self.get_clear_placeholder_url(placeholder)

        with self.login_user_context(superuser):
            data = {'test': 0}
            self.client.post(uri, data)

        self.check.assert_called_once_with(placeholder, superuser)



from cms.test_utils.project.backwards_wizards.wizards import wizard
from cms.wizards.wizard_pool import wizard_pool

# NOTE: We keep this line separate from the actual wizard definition
# because if both are in the same file then importing the wizard causes
# this line to run, which makes it impossible to test properly
wizard_pool.register(wizard)



from cms.test_utils.project.sampleapp.forms import SampleWizardForm
from cms.wizards.wizard_base import Wizard


class SampleWizard(Wizard):
    pass


sample_wizard = SampleWizard(
    title="Sample",
    weight=105,
    form=SampleWizardForm,
    description="Create a new Sample",
)



from django.core.serializers.json import DjangoJSONEncoder
from django.utils.encoding import force_str
from django.utils.functional import Promise
from django.utils.html import conditional_escape


class SafeJSONEncoder(DjangoJSONEncoder):
    def _recursive_escape(self, o, esc=conditional_escape):
        if isinstance(o, dict):
            return type(o)((esc(k), self._recursive_escape(v)) for (k, v) in dict.items(o))
        if isinstance(o, (list, tuple)):
            return type(o)(self._recursive_escape(v) for v in o)
        if type(o) is bool:
            return o
        try:
            return type(o)(esc(o))
        except (ValueError, TypeError):
            return self.default(o)

    def encode(self, o):
        value = self._recursive_escape(o)
        return super().encode(value)

    def default(self, o):
        if isinstance(o, Promise):
            return force_str(o)
        return super().default(o)



from django.conf import settings
from django.urls import include, re_path

from cms.apphook_pool import apphook_pool
from cms.views import details

if settings.APPEND_SLASH:
    reg = re_path(r'^(?P<slug>[0-9A-Za-z-_.//]+)/$', details, name='pages-details-by-slug')
else:
    reg = re_path(r'^(?P<slug>[0-9A-Za-z-_.//]+)$', details, name='pages-details-by-slug')

urlpatterns = [
    # Public pages
    re_path(r'^example/', include('cms.test_utils.project.sampleapp.urls_example', namespace="example1")),
    re_path(r'^example2/', include('cms.test_utils.project.sampleapp.urls_example', namespace="example2")),
    re_path(r'^$', details, {'slug': ''}, name='pages-root'),
    reg,
]

if apphook_pool.get_apphooks():
    """If there are some application urls, add special resolver, so we will
    have standard reverse support.
    """
    from cms.appresolver import get_app_patterns
    urlpatterns = get_app_patterns() + urlpatterns



import copy

from django.contrib.sites.models import Site
from django.test.utils import override_settings

from cms.api import add_plugin, create_page, create_page_content
from cms.exceptions import LanguageError
from cms.forms.utils import update_site_and_page_choices
from cms.models import EmptyPageContent, PageContent
from cms.test_utils.testcases import CMSTestCase
from cms.utils.conf import get_cms_setting, get_languages
from menus.menu_pool import menu_pool

TEMPLATE_NAME = 'tests/rendering/base.html'


def get_primary_language(current_site=None):
    """Fetch the first language of the current site settings."""
    current_site = current_site or Site.objects.get_current()
    return get_languages()[current_site.id][0]['code']


def get_secondary_language(current_site=None):
    """Fetch the other language of the current site settings."""
    current_site = current_site or Site.objects.get_current()
    return get_languages()[current_site.id][1]['code']


@override_settings(
    CMS_TEMPLATES=[
        (TEMPLATE_NAME, TEMPLATE_NAME),
        ('extra_context.html', 'extra_context.html'),
        ('nav_playground.html', 'nav_playground.html'),
    ],
)
class MultilingualTestCase(CMSTestCase):

    def test_create_page(self):
        """
        Test that a page can be created
        and that a new language can be created afterwards in the admin pages
        """

        # Create a new page

        # Use the very first language in the list of languages
        # for the current site
        current_site = Site.objects.get_current()
        TESTLANG = get_primary_language(current_site=current_site)
        page_data = self.get_new_page_data_dbfields(
            site=current_site,
            language=TESTLANG
        )

        page = create_page(**page_data)
        content = page.get_content_obj()

        # A title is set?
        self.assertTrue(bool(content))

        # Has correct title and slug after calling save()?
        self.assertEqual(page.get_title(), page_data['title'])
        self.assertEqual(page.get_slug(TESTLANG), page_data['slug'])
        self.assertEqual(page.get_placeholders(TESTLANG).count(), 2)

        # Do stuff using admin pages
        superuser = self.get_superuser()
        with self.login_user_context(superuser):
            page_data = self.get_pagedata_from_dbfields(page_data)

            self.client.post(self.get_page_change_uri(TESTLANG, page), page_data)

            # Create a different language using the edit admin page
            # This test case is bound in actual experience...
            # pull#1604
            TESTLANG2 = get_secondary_language(current_site=current_site)
            page_data2 = page_data.copy()
            page_data2['title'] = 'ein Titel'
            page_data2['slug'] = 'ein-slug'
            page_data2['cms_page'] = page.pk
            page_data2['language'] = TESTLANG2

            # Ensure that the language version is not returned
            # since it does not exist
            self.assertTrue(isinstance(page.get_content_obj(language=TESTLANG2, fallback=False), EmptyPageContent))

            # Now create it
            self.client.post(self.get_page_add_uri(TESTLANG2, page), page_data2)

            page = page.reload()

            # Test the new language version
            page._clear_internal_cache()
            self.assertEqual(page.get_title(language=TESTLANG2), page_data2['title'])
            self.assertEqual(page.get_slug(language=TESTLANG2), page_data2['slug'])

            # Test the default language version (TESTLANG)
            self.assertEqual(page.get_slug(TESTLANG), page_data['slug'])
            self.assertEqual(page.get_title(language=TESTLANG, fallback=False), page_data['title'])

    def test_multilingual_page(self):
        TESTLANG = get_primary_language()
        TESTLANG2 = get_secondary_language()
        page = create_page("mlpage", "nav_playground.html", TESTLANG)
        create_page_content(TESTLANG2, page.get_title(), page, slug=page.get_slug(TESTLANG))
        page.rescan_placeholders(TESTLANG)
        page = self.reload(page)
        page_placeholder_lang_1 = page.get_placeholders(TESTLANG)[0]
        page_placeholder_lang_2 = page.get_placeholders(TESTLANG2)[0]
        add_plugin(page_placeholder_lang_1, "TextPlugin", TESTLANG, body="test")
        add_plugin(page_placeholder_lang_2, "TextPlugin", TESTLANG2, body="test")
        self.assertEqual(page_placeholder_lang_1.get_plugins(language=TESTLANG).count(), 1)
        self.assertEqual(page_placeholder_lang_2.get_plugins(language=TESTLANG2).count(), 1)

    def test_hide_untranslated(self):
        TESTLANG = get_primary_language()
        TESTLANG2 = get_secondary_language()
        page = create_page("mlpage-%s" % TESTLANG, "nav_playground.html", TESTLANG)
        create_page_content(TESTLANG2, "mlpage-%s" % TESTLANG2, page, slug=page.get_slug(TESTLANG))
        create_page("mlpage-2-%s" % TESTLANG, "nav_playground.html", TESTLANG, parent=page)

        lang_settings = copy.deepcopy(get_cms_setting('LANGUAGES'))

        request_1 = self.get_request('/%s/' % TESTLANG, TESTLANG)
        request_2 = self.get_request('/%s/' % TESTLANG2, TESTLANG2)

        request_1_menu_renderer = menu_pool.get_renderer(request_1)
        request_2_menu_renderer = menu_pool.get_renderer(request_2)

        lang_settings[1][1]['hide_untranslated'] = False
        with self.settings(CMS_LANGUAGES=lang_settings):
            request_1_nodes = request_1_menu_renderer.get_menu('CMSMenu').get_nodes(request_1)
            request_2_nodes = request_2_menu_renderer.get_menu('CMSMenu').get_nodes(request_2)
            list_1 = [node.id for node in request_1_nodes]
            list_2 = [node.id for node in request_2_nodes]
            self.assertEqual(list_1, list_2)
            self.assertEqual(len(list_1), 2)

        lang_settings[1][1]['hide_untranslated'] = True
        with self.settings(CMS_LANGUAGES=lang_settings):
            request_1_nodes = request_1_menu_renderer.get_menu('CMSMenu').get_nodes(request_1)
            request_2_nodes = request_2_menu_renderer.get_menu('CMSMenu').get_nodes(request_2)
            list_1 = [node.id for node in request_1_nodes]
            list_2 = [node.id for node in request_2_nodes]
            self.assertNotEqual(list_1, list_2)
            self.assertEqual(len(list_2), 1)
            self.assertEqual(len(list_1), 2)

    def test_frontend_lang(self):
        superuser = self.get_superuser()
        lang_settings = copy.deepcopy(get_cms_setting('LANGUAGES'))
        lang_settings[1][0]['public'] = False
        with self.settings(CMS_LANGUAGES=lang_settings, LANGUAGE_CODE="en"):
            page = create_page("page1", "nav_playground.html", "en")
            create_page_content("de", page.get_title(), page, slug=page.get_slug('en'))
            page2 = create_page("page2", "nav_playground.html", "en")
            create_page_content("de", page2.get_title(), page2, slug=page2.get_slug('en'))
            page3 = create_page("page2", "nav_playground.html", "en")
            create_page_content("de", page3.get_title(), page3, slug=page3.get_slug('en'))
            create_page("page4", "nav_playground.html", "de")

            page.set_as_homepage()

            # The "en" language is set to public -> False.
            # Because the request is to the root (homepage),
            # the page is redirected to the default language's homepage
            response = self.client.get("/en/")
            self.assertRedirects(response, '/de/')

            # Authenticated requests to a private language
            # will render the page normally as long as the language
            # is available on the page
            with self.login_user_context(superuser):
                response = self.client.get("/en/")
                self.assertEqual(response.status_code, 200)

            response = self.client.get("/en/page2/")
            self.assertEqual(response.status_code, 404)
            response = self.client.get("/de/")
            self.assertEqual(response.status_code, 200)
            response = self.client.get("/de/page2/")
            self.assertEqual(response.status_code, 200)

            # check if the admin can see non-public langs
            with self.login_user_context(self.get_superuser()):
                response = self.client.get("/en/page2/")
                self.assertEqual(response.status_code, 200)
                response = self.client.get("/en/page4/")
                self.assertEqual(response.status_code, 302)
            self.client.logout()
            response = self.client.get("/en/page4/")
            self.assertEqual(response.status_code, 404)

    def test_page_with_invalid_language_for_anon_user(self):
        site_2 = Site.objects.create(id=2, name='example-2.com', domain='example-2.com')
        self.create_homepage(
            "page",
            template='nav_playground.html',
            language="de",
            site=site_2,
        )
        page_2 = create_page(
            "page",
            template='nav_playground.html',
            language="de",
            site=site_2,
        )

        with self.settings(SITE_ID=2, LANGUAGE_CODE='en'):
            # url uses "en" as the request language
            # but the site is configured to use "de" and "fr"
            response = self.client.get('/en/')
            self.assertRedirects(response, '/de/')
            response = self.client.get('/en/%s/' % page_2.get_path('de'))
            self.assertEqual(response.status_code, 404)

    def test_page_with_invalid_language_for_auth_user(self):
        site_2 = Site.objects.create(id=2, name='example-2.com', domain='example-2.com')
        superuser = self.get_superuser()
        self.create_homepage(
            "page",
            template='nav_playground.html',
            language="de",
            site=site_2,
        )
        page_2 = create_page(
            "page",
            template='nav_playground.html',
            language="de",
            site=site_2,
        )

        with self.settings(SITE_ID=2, LANGUAGE_CODE='en'):
            with self.login_user_context(superuser):
                # url uses "en" as the request language
                # but the site is configured to use "de" and "fr"
                response = self.client.get('/en/')
                self.assertRedirects(response, '/de/')
                response = self.client.get('/en/%s/' % page_2.get_path('de'))
                self.assertEqual(response.status_code, 404)

    def test_language_fallback(self):
        """
        Test language fallbacks in details view
        """
        from cms.views import details
        p1 = create_page("page", "nav_playground.html", "en")
        p1.set_as_homepage()

        # There's no "de" translation.
        # Fallbacks are configured.
        # The cms is set to redirect on fallback.
        request = self.get_request('/de/', 'de')
        response = details(request, p1.get_path('en'))
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response['Location'], '/en/')

        # There's no "de" translation.
        # There's no fallbacks configured.
        lang_settings = copy.deepcopy(get_cms_setting('LANGUAGES'))
        lang_settings[1][0]['fallbacks'] = []
        lang_settings[1][1]['fallbacks'] = []

        with self.settings(CMS_LANGUAGES=lang_settings):
            response = self.client.get("/de/")
            self.assertEqual(response.status_code, 404)

        # There's no "de" translation.
        # Fallbacks are configured.
        # The cms is set to render in place instead of redirecting
        # to the fallback.
        lang_settings = copy.deepcopy(get_cms_setting('LANGUAGES'))
        lang_settings[1][0]['redirect_on_fallback'] = False
        lang_settings[1][1]['redirect_on_fallback'] = False

        with self.settings(CMS_LANGUAGES=lang_settings):
            response = self.client.get("/de/")
            self.assertRedirects(response, '/en/')

    def test_no_english_defined(self):
        with self.settings(
            TEMPLATE_CONTEXT_PROCESSORS=[],
            CMS_LANGUAGES={
                1: [
                    {'code': 'de', 'name': 'German', 'public': True, 'fallbacks': []},
                ]},
        ):
            try:
                update_site_and_page_choices(language='en-us')
            except LanguageError:
                self.fail("LanguageError raised")

    def test_wrong_plugin_language(self):
        page = create_page("page", "nav_playground.html", "en")
        ph_en = page.get_placeholders("en").get(slot="body")
        add_plugin(ph_en, "TextPlugin", "en", body="I'm the first")
        title = PageContent(title="page", language="ru", page=page)
        title.save()
        # add wrong plugin language
        add_plugin(ph_en, "TextPlugin", "ru", body="I'm the second")
        endpoint = page.get_absolute_url()
        superuser = self.get_superuser()
        with self.login_user_context(superuser):
            response = self.client.get(endpoint)
            self.assertEqual(response.status_code, 200)



from django.conf import settings
from django.urls import include, re_path

from cms import views
from cms.apphook_pool import apphook_pool
from cms.appresolver import get_app_patterns
from cms.constants import SLUG_REGEXP

if settings.APPEND_SLASH:
    regexp = r'^(?P<slug>%s)/$' % SLUG_REGEXP
else:
    regexp = r'^(?P<slug>%s)$' % SLUG_REGEXP

if apphook_pool.get_apphooks():
    # If there are some application urls, use special resolver,
    # so we will have standard reverse support.
    urlpatterns = get_app_patterns()
else:
    urlpatterns = []


urlpatterns.extend([
    re_path(r'^cms_login/$', views.login, name='cms_login'),
    re_path(r'^cms_wizard/', include('cms.wizards.urls')),
    re_path(regexp, views.details, name='pages-details-by-slug'),
    re_path(r'^$', views.details, {'slug': ''}, name='pages-root'),
])



"""
This module manages placeholder caching. We use a cache-versioning strategy
in which each (placeholder x lang x site_id) manages its own version. The
actual cache includes additional keys appropriate for the placeholders
get_vary_cache_on().

Invalidation of a placeholder's cache simply increments the version number for
the (placeholder x lang) pair, which renders any cache entries for that
placeholder under that version inaccessible. Those cache entries will simply
expire and will be purged according to the policy of the cache backend in-use.

The cache entries themselves may include additional sub-keys, according to the
list of VARY header-names as returned by placeholder.get_vary_cache_on() and
the current HTTPRequest object.

The vary-on header-names are also stored with the version. This enables us to
check for cache hits without re-computing placeholder.get_vary_cache_on().
"""
import hashlib
import time

from django.utils.timezone import now

from cms.utils.conf import get_cms_setting
from cms.utils.helpers import get_header_name, get_timezone_name


def _get_placeholder_cache_version_key(placeholder, lang, site_id):
    """
    Returns the version key for the given «placeholder», «lang» and «site_id».

    Invalidating this (via clear_placeholder_cache by replacing the stored
    value with a new value) will effectively make all "sub-caches" relating to
    this (placeholder x lang) inaccessible. Sub-caches include caches per TZ
    and per VARY header.
    """
    prefix = get_cms_setting('CACHE_PREFIX')
    key = '{prefix}|placeholder_cache_version|id:{id}|lang:{lang}|site:{site}'.format(
        prefix=prefix,
        id=placeholder.pk,
        lang=str(lang),
        site=site_id,
    )
    if len(key) > 250:
        key = '{prefix}|{hash}'.format(
            prefix=prefix,
            hash=hashlib.sha1(key.encode('utf-8')).hexdigest(),
        )
    return key


def _get_placeholder_cache_version(placeholder, lang, site_id):
    """
    Gets the (placeholder x lang)'s current version and vary-on header-names
    list, if present, otherwise resets to («timestamp», []).
    """
    from django.core.cache import cache

    key = _get_placeholder_cache_version_key(placeholder, lang, site_id)
    cached = cache.get(key)
    if cached:
        version, vary_on_list = cached
    else:
        version = int(time.time() * 1000000)
        vary_on_list = []
        _set_placeholder_cache_version(placeholder, lang, site_id, version, vary_on_list)
    return version, vary_on_list


def _set_placeholder_cache_version(placeholder, lang, site_id, version, vary_on_list=None, duration=None):
    """
    Sets the (placeholder x lang)'s version and vary-on header-names list.
    """
    from django.core.cache import cache

    key = _get_placeholder_cache_version_key(placeholder, lang, site_id)

    if not version or version < 1:
        version = int(time.time() * 1000000)

    if vary_on_list is None:
        vary_on_list = []

    cache.set(key, (version, vary_on_list), duration)


def _get_placeholder_cache_key(placeholder, lang, site_id, request, soft=False):
    """
    Returns the fully-addressed cache key for the given placeholder and
    the request.

    The kwarg «soft» should be set to True if getting the cache key to then
    read from the cache. If instead the key retrieval is to support a cache
    write, let «soft» be False.
    """
    prefix = get_cms_setting('CACHE_PREFIX')
    version, vary_on_list = _get_placeholder_cache_version(placeholder, lang, site_id)
    main_key = '{prefix}|render_placeholder|id:{id}|lang:{lang}|site:{site}|tz:{tz}|v:{version}'.format(
        prefix=prefix,
        id=placeholder.pk,
        lang=lang,
        site=site_id,
        tz=get_timezone_name(),
        version=version,
    )

    if not soft:
        # We are about to write to the cache, so we want to get the latest
        # vary_cache_on headers and the correct cache expiration, ignoring any
        # we already have. If the placeholder has already been rendered, this
        # will be very efficient (zero-additional queries) due to the caching
        # of all its plugins during the rendering process anyway.
        vary_on_list = placeholder.get_vary_cache_on(request)
        duration = placeholder.get_cache_expiration(request, now())
        # Update the main placeholder cache version
        _set_placeholder_cache_version(
            placeholder, lang, site_id, version, vary_on_list, duration)

    sub_key_list = []
    for key in vary_on_list:
        value = request.META.get(get_header_name(key)) or '_'
        sub_key_list.append(key + ':' + value)

    cache_key = main_key
    if sub_key_list:
        cache_key += '|' + '|'.join(sub_key_list)

    if len(cache_key) > 250:
        cache_key = '{prefix}|{hash}'.format(
            prefix=prefix,
            hash=hashlib.sha1(cache_key.encode('utf-8')).hexdigest(),
        )

    return cache_key


def set_placeholder_cache(placeholder, lang, site_id, content, request):
    """
    Sets the (correct) placeholder cache with the rendered placeholder.
    """
    from django.core.cache import cache

    key = _get_placeholder_cache_key(placeholder, lang, site_id, request)

    duration = min(
        get_cms_setting('CACHE_DURATIONS')['content'],
        placeholder.get_cache_expiration(request, now())
    )
    cache.set(key, content, duration)
    # "touch" the cache-version, so that it stays as fresh as this content.
    version, vary_on_list = _get_placeholder_cache_version(placeholder, lang, site_id)
    _set_placeholder_cache_version(
        placeholder, lang, site_id, version, vary_on_list, duration=duration
    )


def get_placeholder_cache(placeholder, lang, site_id, request):
    """
    Returns the placeholder from cache respecting the placeholder's
    VARY headers.
    """
    from django.core.cache import cache

    key = _get_placeholder_cache_key(placeholder, lang, site_id, request, soft=True)
    content = cache.get(key)
    return content


def clear_placeholder_cache(placeholder, lang, site_id):
    """
    Invalidates all existing cache entries for (placeholder x lang x site_id).
    We don't need to re-store the vary_on_list, because the cache is now
    effectively empty.
    """
    version = int(time.time() * 1000000)
    _set_placeholder_cache_version(placeholder, lang, site_id, version, [])



import os

from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.core.files.storage import FileSystemStorage
from django.forms import Form
from django.template.response import SimpleTemplateResponse
from django.urls import NoReverseMatch
from formtools.wizard.views import SessionWizardView

from cms.models import Page
from cms.utils import get_current_site
from cms.utils.i18n import get_site_language_from_request

from .forms import WizardStep1Form, WizardStep2BaseForm, step2_form_factory
from .wizard_pool import wizard_pool


class WizardCreateView(SessionWizardView):
    template_name = 'cms/wizards/start.html'
    file_storage = FileSystemStorage(
        location=os.path.join(settings.MEDIA_ROOT, 'wizard_tmp_files'))

    form_list = [
        ('0', WizardStep1Form),
        # Form is used as a placeholder form.
        # the real form will be loaded after step 0
        ('1', Form),
    ]

    def dispatch(self, *args, **kwargs):
        user = self.request.user

        if not user.is_active or not user.is_staff:
            raise PermissionDenied
        self.site = get_current_site()
        return super().dispatch(*args, **kwargs)

    def get_current_step(self):
        """Returns the current step, if possible, else None."""
        try:
            return self.steps.current
        except AttributeError:
            return None

    def is_first_step(self, step=None):
        step = step or self.get_current_step()
        return step == '0'

    def is_second_step(self, step=None):
        step = step or self.get_current_step()
        return step == '1'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        if self.is_second_step():
            context['wizard_entry'] = self.get_selected_entry()
        return context

    def get_form(self, step=None, data=None, files=None):
        if step is None:
            step = self.steps.current

        # We need to grab the page from pre-validated data so that the wizard
        # has it to prepare the list of valid entries.
        if data:
            page_key = "{0}-page".format(step)
            self.page_pk = data.get(page_key, None)
        else:
            self.page_pk = None

        if self.is_second_step(step):
            self.form_list[step] = self.get_step_2_form(step, data, files)
        return super().get_form(step, data, files)

    def get_form_kwargs(self, step=None):
        """This is called by self.get_form()"""
        kwargs = super().get_form_kwargs()
        kwargs['wizard_request'] = self.request
        kwargs['wizard_site'] = self.site
        if self.is_second_step(step):
            kwargs['wizard_page'] = self.get_origin_page()
            kwargs['wizard_language'] = self.get_origin_language()
        else:
            page_pk = self.page_pk or self.request.GET.get('page', None)
            if page_pk and page_pk != 'None':
                kwargs['wizard_page'] = Page.objects.filter(pk=page_pk).first()
            else:
                kwargs['wizard_page'] = None
            kwargs['wizard_language'] = get_site_language_from_request(
                self.request,
                site_id=self.site.pk,
            )
        return kwargs

    def get_form_initial(self, step):
        """This is called by self.get_form()"""
        initial = super().get_form_initial(step)
        if self.is_first_step(step):
            initial['page'] = self.request.GET.get('page')
            initial['language'] = self.request.GET.get('language')
        return initial

    def get_step_2_form(self, step=None, data=None, files=None):
        entry_form_class = self.get_selected_entry().form
        step_2_base_form = self.get_step_2_base_form()

        form = step2_form_factory(
            mixin_cls=step_2_base_form,
            entry_form_class=entry_form_class,
        )
        return form

    def get_step_2_base_form(self):
        """
        Returns the base form to be used for step 2.
        This form is sub classed dynamically by the form defined per module.
        """
        return WizardStep2BaseForm

    def get_template_names(self):
        if self.is_first_step():
            template_name = self.template_name
        else:
            template_name = self.get_selected_entry().template_name
        return template_name

    def done(self, form_list, **kwargs):
        """
        This step only runs if all forms are valid. Simply emits a simple
        template that uses JS to redirect to the newly created object.
        """
        form_one, form_two = list(form_list)
        instance = form_two.save()
        url = self.get_success_url(instance)
        language = form_one.cleaned_data['language']
        if not url:
            page = self.get_origin_page()
            if page:
                try:
                    url = page.get_absolute_url(language)
                except NoReverseMatch:
                    url = '/'
            else:
                url = '/'

        return SimpleTemplateResponse("cms/wizards/done.html", {"url": url})

    def get_selected_entry(self):
        data = self.get_cleaned_data_for_step('0')
        return wizard_pool.get_entry(data['entry'])

    def get_origin_page(self):
        data = self.get_cleaned_data_for_step('0')
        return data.get('page')

    def get_origin_language(self):
        data = self.get_cleaned_data_for_step('0')
        return data.get('language')

    def get_success_url(self, instance):
        entry = self.get_selected_entry()
        language = self.get_origin_language()
        success_url = entry.get_success_url(
            obj=instance,
            language=language,
        )
        return success_url



import uuid
import warnings
from urllib.parse import parse_qsl, urlparse

from django import forms
from django.contrib import admin
from django.contrib.admin.helpers import AdminForm
from django.contrib.admin.utils import get_deleted_objects
from django.core.exceptions import PermissionDenied
from django.db import transaction
from django.http import (
    HttpResponse,
    HttpResponseBadRequest,
    HttpResponseForbidden,
    HttpResponseNotFound,
    HttpResponseRedirect,
)
from django.shortcuts import get_list_or_404, get_object_or_404, render
from django.template.response import TemplateResponse
from django.urls import re_path
from django.utils.decorators import method_decorator
from django.utils.encoding import force_str
from django.utils.html import conditional_escape
from django.utils.translation import get_language_from_path
from django.utils.translation import gettext as _
from django.views.decorators.clickjacking import xframe_options_sameorigin
from django.views.decorators.http import require_POST

from cms import operations
from cms.admin.forms import PluginAddValidationForm
from cms.constants import SLUG_REGEXP
from cms.exceptions import PluginLimitReached
from cms.models.placeholdermodel import Placeholder
from cms.models.placeholderpluginmodel import PlaceholderReference
from cms.models.pluginmodel import CMSPlugin
from cms.plugin_pool import plugin_pool
from cms.signals import post_placeholder_operation, pre_placeholder_operation
from cms.toolbar.utils import get_plugin_tree_as_json
from cms.utils import get_current_site
from cms.utils.conf import get_cms_setting
from cms.utils.i18n import get_language_code, get_language_list
from cms.utils.plugins import (
    copy_plugins_to_placeholder,
    has_reached_plugin_limit,
)
from cms.utils.urlutils import admin_reverse
from cms.views import (
    render_object_edit,
    render_object_preview,
    render_object_structure,
)

_no_default = object()


def get_int(int_str, default=_no_default):
    """
    For convenience a get-like method for taking the int() of a string.
    :param int_str: the string to convert to integer
    :param default: an optional value to return if ValueError is raised.
    :return: the int() of «int_str» or «default» on exception.
    """
    if default == _no_default:
        return int(int_str)
    else:
        try:
            return int(int_str)
        except ValueError:
            return default


def _instance_overrides_method(base, instance, method_name):
    """
    Returns True if instance overrides a method (method_name)
    inherited from base.
    """
    bound_method = getattr(instance.__class__, method_name)
    unbound_method = getattr(base, method_name)
    return unbound_method != bound_method


class FrontendEditableAdminMixin():
    """
    Adding ``FrontendEditableAdminMixin`` to  models admin class allows to open that admin
    in the frontend by double-clicking on fields rendered with the ``render_model`` template
    tag.
    """
    frontend_editable_fields = []

    def get_urls(self):
        """
        Register the url for the single field edit view
        """
        info = "%s_%s" % (self.model._meta.app_label, self.model._meta.model_name)
        def pat(regex, fn):
            return re_path(regex, self.admin_site.admin_view(fn), name="%s_%s" % (info, fn.__name__))
        url_patterns = [
            pat(r'edit-field/(%s)/([a-z\-]+)/$' % SLUG_REGEXP, self.edit_field),
        ]
        return url_patterns + super().get_urls()

    def _get_object_for_single_field(self, object_id, language):
        # Quick and dirty way to retrieve objects for django-hvad
        # Cleaner implementation will extend this method in a child mixin
        try:
            return self.model.objects.language(language).get(pk=object_id)
        except AttributeError:
            return self.model.objects.get(pk=object_id)

    def edit_field(self, request, object_id, language):
        obj = self._get_object_for_single_field(object_id, language)
        opts = obj.__class__._meta
        saved_successfully = False
        cancel_clicked = request.POST.get("_cancel", False)
        raw_fields = request.GET.get("edit_fields")
        fields = [field for field in raw_fields.split(",") if field in self.frontend_editable_fields]
        if not fields:
            context = {
                'opts': opts,
                'message': force_str(_("Field %s not found")) % raw_fields
            }
            return render(request, 'admin/cms/page/plugin/error_form.html', context)
        if not request.user.has_perm("{0}.change_{1}".format(self.model._meta.app_label,
                                                             self.model._meta.model_name)):
            context = {
                'opts': opts,
                'message': force_str(_("You do not have permission to edit this item"))
            }
            return render(request, 'admin/cms/page/plugin/error_form.html', context)
            # Dynamically creates the form class with only `field_name` field
        # enabled
        form_class = self.get_form(request, obj, fields=fields)
        if not cancel_clicked and request.method == 'POST':
            form = form_class(instance=obj, data=request.POST)
            if form.is_valid():
                form.save()
                saved_successfully = True
        else:
            form = form_class(instance=obj)
        admin_form = AdminForm(form, fieldsets=[(None, {'fields': fields})], prepopulated_fields={},
                               model_admin=self)
        media = self.media + admin_form.media
        context = {
            'CMS_MEDIA_URL': get_cms_setting('MEDIA_URL'),
            'title': opts.verbose_name,
            'plugin': None,
            'plugin_id': None,
            'adminform': admin_form,
            'add': False,
            'is_popup': True,
            'media': media,
            'opts': opts,
            'change': True,
            'save_as': False,
            'has_add_permission': False,
            'window_close_timeout': 10,
        }
        if cancel_clicked:
            # cancel button was clicked
            context.update({
                'cancel': True,
            })
            return render(request, 'admin/cms/page/plugin/confirm_form.html', context)
        if not cancel_clicked and request.method == 'POST' and saved_successfully:
            return render(request, 'admin/cms/page/plugin/confirm_form.html', context)
        return render(request, 'admin/cms/page/plugin/change_form.html', context)


class PlaceholderAdminMixinBase(forms.MediaDefiningClass):
    def __new__(cls, name, bases, attrs):
        super_new = super().__new__
        parents = [b for b in bases if isinstance(b, PlaceholderAdminMixinBase)]

        if not parents:
            return super_new(cls, name, bases, attrs)
        warnings.warn(
            "PlaceholderAdminMixin is no longer needed and thus will be removed in django CMS 5.0",
            DeprecationWarning,
            stacklevel=2,
        )
        return super_new(cls, name, bases, attrs)


class PlaceholderAdminMixin(metaclass=PlaceholderAdminMixinBase):
    """
    .. warning::

        PlaceholderAdminMixin is deprecated. It is no longer needed and thus will be removed
    """
    pass


class PlaceholderAdmin(admin.ModelAdmin):

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False

    def has_module_permission(self, request):
        return False

    def get_urls(self):
        """
        Register the plugin specific urls (add/edit/copy/remove/move)
        """
        info = "%s_%s" % (self.model._meta.app_label, self.model._meta.model_name)
        def pat(regex, fn):
            return re_path(regex, self.admin_site.admin_view(fn), name="%s_%s" % (info, fn.__name__))
        url_patterns = [
            pat(r'^copy-plugins/$', self.copy_plugins),
            pat(r'^add-plugin/$', self.add_plugin),
            pat(r'^edit-plugin/([0-9]+)/$', self.edit_plugin),
            pat(r'^delete-plugin/([0-9]+)/$', self.delete_plugin),
            pat(r'^clear-placeholder/([0-9]+)/$', self.clear_placeholder),
            pat(r'^move-plugin/$', self.move_plugin),
            # Register object edit/structure/preview endpoints.
            # pat(r'^object/(?P<content_type_id>\d+)/structure/(?P<object_id>.+)$', render_object_structure),
            pat(r'^object/([0-9]+)/edit/([0-9]+)/$', render_object_edit),
            pat(r'^object/([0-9]+)/structure/([0-9]+)/$', render_object_structure),
            pat(r'^object/([0-9]+)/preview/([0-9]+)/$', render_object_preview),
        ]
        return url_patterns

    def _get_operation_language(self, request):
        # Unfortunately the ?language GET query
        # has a special meaning on the CMS.
        # It allows users to see another language while maintaining
        # the same url. This complicates language detection.
        site = get_current_site()
        parsed_url = urlparse(request.GET['cms_path'])
        queries = dict(parse_qsl(parsed_url.query))
        language = queries.get('language')

        if not language:
            language = get_language_from_path(parsed_url.path)
        return get_language_code(language, site_id=site.pk)

    def _get_operation_origin(self, request):
        return urlparse(request.GET['cms_path']).path

    def _send_pre_placeholder_operation(self, request, operation, **kwargs):
        token = str(uuid.uuid4())

        if not request.GET.get('cms_path'):
            warnings.warn('All custom placeholder admin endpoints require '
                          'a "cms_path" GET query which points to the path '
                          'where the request originates from.'
                          'This backwards compatible shim will be removed on 3.5 '
                          'and an HttpBadRequest response will be returned instead.',
                          UserWarning)
            return token

        pre_placeholder_operation.send(
            sender=self.__class__,
            operation=operation,
            request=request,
            language=self._get_operation_language(request),
            token=token,
            origin=self._get_operation_origin(request),
            **kwargs
        )
        return token

    def _send_post_placeholder_operation(self, request, operation, token, **kwargs):
        if not request.GET.get('cms_path'):
            # No need to re-raise the warning
            return

        post_placeholder_operation.send(
            sender=self.__class__,
            operation=operation,
            request=request,
            language=self._get_operation_language(request),
            token=token,
            origin=self._get_operation_origin(request),
            **kwargs
        )

    def _get_plugin_from_id(self, plugin_id):
        queryset = CMSPlugin.objects.values_list('plugin_type', flat=True)
        plugin_type = get_list_or_404(queryset, pk=plugin_id)[0]
        # CMSPluginBase subclass
        plugin_class = plugin_pool.get_plugin(plugin_type)
        real_queryset = plugin_class.get_render_queryset().select_related('parent', 'placeholder')
        return get_object_or_404(real_queryset, pk=plugin_id)

    def has_add_plugin_permission(self, request, placeholder, plugin_type):
        return placeholder.has_add_plugin_permission(request.user, plugin_type)

    def has_change_plugin_permission(self, request, plugin):
        placeholder = plugin.placeholder
        return placeholder.has_change_plugin_permission(request.user, plugin)

    def has_delete_plugin_permission(self, request, plugin):
        placeholder = plugin.placeholder
        return placeholder.has_delete_plugin_permission(request.user, plugin)

    def has_copy_plugins_permission(self, request, plugins):
        # Plugins can only be copied to the clipboard
        placeholder = request.toolbar.clipboard
        return placeholder.has_add_plugins_permission(request.user, plugins)

    def has_copy_from_clipboard_permission(self, request, placeholder, plugins):
        return placeholder.has_add_plugins_permission(request.user, plugins)

    def has_copy_from_placeholder_permission(self, request, source_placeholder, target_placeholder, plugins):
        if not source_placeholder.has_add_plugins_permission(request.user, plugins):
            return False
        return target_placeholder.has_add_plugins_permission(request.user, plugins)

    def has_move_plugin_permission(self, request, plugin, target_placeholder):
        placeholder = plugin.placeholder
        return placeholder.has_move_plugin_permission(request.user, plugin, target_placeholder)

    def has_clear_placeholder_permission(self, request, placeholder, language=None):
        if language:
            languages = [language]
        else:
            # fetch all languages this placeholder contains
            # based on it's plugins
            languages = (
                placeholder
                .cmsplugin_set
                .values_list('language', flat=True)
                .distinct()
                .order_by()
            )
        return placeholder.has_clear_permission(request.user, languages)

    def get_placeholder_template(self, request, placeholder):
        pass

    @xframe_options_sameorigin
    def add_plugin(self, request):
        """
        Shows the add plugin form and saves it on POST.

        Requires the following GET parameters:
            - cms_path
            - placeholder_id
            - plugin_type
            - plugin_language
            - plugin_position
            - plugin_parent (optional)
        """
        form = PluginAddValidationForm(request.GET)

        if not form.is_valid():
            # list() is necessary for python 3 compatibility.
            # errors is s dict mapping fields to a list of errors
            # for that field.
            error = list(form.errors.values())[0][0]
            return HttpResponseBadRequest(conditional_escape(force_str(error)))

        plugin_data = form.cleaned_data
        placeholder = plugin_data['placeholder_id']
        plugin_type = plugin_data['plugin_type']

        if not self.has_add_plugin_permission(request, placeholder, plugin_type):
            message = force_str(_('You do not have permission to add a plugin'))
            return HttpResponseForbidden(message)

        if not placeholder.check_source(request.user):
            message = force_str(_('You do not have permission to add a plugin'))
            return HttpResponseForbidden(message)

        plugin_class = plugin_pool.get_plugin(plugin_type)
        plugin_instance = plugin_class(plugin_class.model, self.admin_site)

        # Setting attributes on the form class is perfectly fine.
        # The form class is created by modelform factory every time
        # this get_form() method is called.
        plugin_instance._cms_initial_attributes = {
            'language': plugin_data['plugin_language'],
            'placeholder': plugin_data['placeholder_id'],
            'parent': plugin_data.get('plugin_parent', None),
            'plugin_type': plugin_data['plugin_type'],
            'position': plugin_data['plugin_position'],
        }

        response = plugin_instance.add_view(request)

        plugin = getattr(plugin_instance, 'saved_object', None)

        if plugin_instance._operation_token:
            self._send_post_placeholder_operation(
                request,
                operation=operations.ADD_PLUGIN,
                token=plugin_instance._operation_token,
                plugin=plugin,
                placeholder=plugin.placeholder,
            )
        return response

    @method_decorator(require_POST)
    @xframe_options_sameorigin
    @transaction.atomic
    def copy_plugins(self, request):
        """
        POST request should have the following data:

        - cms_path
        - source_language
        - source_placeholder_id
        - source_plugin_id (optional)
        - target_language
        - target_placeholder_id
        """
        source_placeholder_id = request.POST['source_placeholder_id']
        target_language = request.POST['target_language']
        target_placeholder_id = request.POST['target_placeholder_id']
        source_placeholder = get_object_or_404(Placeholder, pk=source_placeholder_id)
        target_placeholder = get_object_or_404(Placeholder, pk=target_placeholder_id)

        if not target_language or target_language not in get_language_list():
            return HttpResponseBadRequest(force_str(_("Language must be set to a supported language!")))

        copy_to_clipboard = target_placeholder.pk == request.toolbar.clipboard.pk
        source_plugin_id = request.POST.get('source_plugin_id', None)

        if copy_to_clipboard and source_plugin_id:
            new_plugin = self._copy_plugin_to_clipboard(
                request,
                target_placeholder,
            )
            new_plugins = [new_plugin]
        elif copy_to_clipboard:
            new_plugin = self._copy_placeholder_to_clipboard(
                request,
                source_placeholder,
                target_placeholder,
            )
            new_plugins = [new_plugin]
        else:
            new_plugins = self._add_plugins_from_placeholder(
                request,
                source_placeholder,
                target_placeholder,
            )
        data = get_plugin_tree_as_json(request, new_plugins)
        return HttpResponse(data, content_type='application/json')

    def _copy_plugin_to_clipboard(self, request, target_placeholder):
        source_language = request.POST['source_language']
        source_plugin_id = request.POST.get('source_plugin_id')
        target_language = request.POST['target_language']

        source_plugin = get_object_or_404(
            CMSPlugin,
            pk=source_plugin_id,
            language=source_language,
        )
        old_plugins = [source_plugin] + list(source_plugin.get_descendants())

        if not self.has_copy_plugins_permission(request, old_plugins):
            message = _('You do not have permission to copy these plugins.')
            raise PermissionDenied(force_str(message))

        if not target_placeholder.check_source(request.user):
            message = _('You do not have permission to copy these plugins.')
            raise PermissionDenied(message)

        # Empty the clipboard
        target_placeholder.clear()

        copied_plugins = copy_plugins_to_placeholder(
            old_plugins,
            placeholder=target_placeholder,
            language=target_language,
        )
        return copied_plugins[0]

    def _copy_placeholder_to_clipboard(self, request, source_placeholder, target_placeholder):
        source_language = request.POST['source_language']
        target_language = request.POST['target_language']

        # User is copying the whole placeholder to the clipboard.
        old_plugins = source_placeholder.get_plugins_list(language=source_language)

        if not self.has_copy_plugins_permission(request, old_plugins):
            message = _('You do not have permission to copy this placeholder.')
            raise PermissionDenied(force_str(message))

        if not target_placeholder.check_source(request.user):
            message = _('You do not have permission to copy this placeholder.')
            raise PermissionDenied(force_str(message))

        # Empty the clipboard
        target_placeholder.clear()

        # Create a PlaceholderReference plugin which in turn
        # creates a blank placeholder called "clipboard"
        # the real clipboard has the reference placeholder inside but the plugins
        # are inside of the newly created blank clipboard.
        # This allows us to wrap all plugins in the clipboard under one plugin
        reference = PlaceholderReference.objects.create(
            name=source_placeholder.get_label(),
            plugin_type='PlaceholderPlugin',
            language=target_language,
            placeholder=target_placeholder,
        )
        copy_plugins_to_placeholder(
            old_plugins,
            placeholder=reference.placeholder_ref,
            language=target_language,
        )
        return reference

    def _add_plugins_from_placeholder(self, request, source_placeholder, target_placeholder):
        # Plugins are being copied from a placeholder in another language
        # using the "Copy from language" placeholder operation.
        source_language = request.POST['source_language']
        target_language = request.POST['target_language']

        old_plugins = source_placeholder.get_plugins_list(language=source_language)

        # Check if the user can copy plugins from source placeholder to
        # target placeholder.
        has_permissions = self.has_copy_from_placeholder_permission(
            request,
            source_placeholder,
            target_placeholder,
            old_plugins,
        )

        if not has_permissions:
            message = _('You do not have permission to copy these plugins.')
            raise PermissionDenied(force_str(message))

        if not target_placeholder.check_source(request.user):
            message = _('You do not have permission to copy these plugins.')
            raise PermissionDenied(force_str(message))

        target_tree_order = target_placeholder.get_plugin_tree_order(
            language=target_language,
            parent_id=None,
        )

        operation_token = self._send_pre_placeholder_operation(
            request,
            operation=operations.ADD_PLUGINS_FROM_PLACEHOLDER,
            plugins=old_plugins,
            source_language=source_language,
            source_placeholder=source_placeholder,
            target_language=target_language,
            target_placeholder=target_placeholder,
            target_order=target_tree_order,
        )

        copied_plugins = copy_plugins_to_placeholder(old_plugins, target_placeholder, language=target_language)
        new_plugin_ids = (new.pk for new in copied_plugins)

        target_placeholder.clear_cache(target_language)

        new_plugins = CMSPlugin.objects.filter(pk__in=new_plugin_ids)
        new_plugins = list(new_plugins)

        self._send_post_placeholder_operation(
            request,
            operation=operations.ADD_PLUGINS_FROM_PLACEHOLDER,
            token=operation_token,
            plugins=new_plugins,
            source_language=source_language,
            source_placeholder=source_placeholder,
            target_language=target_language,
            target_placeholder=target_placeholder,
            target_order=target_tree_order,
        )
        return new_plugins

    @xframe_options_sameorigin
    def edit_plugin(self, request, plugin_id):
        try:
            plugin_id = int(plugin_id)
        except ValueError:
            return HttpResponseNotFound(force_str(_("Plugin not found")))

        obj = self._get_plugin_from_id(plugin_id)

        # CMSPluginBase subclass instance
        plugin_instance = obj.get_plugin_class_instance(admin=self.admin_site)

        if not self.has_change_plugin_permission(request, obj):
            return HttpResponseForbidden(force_str(_("You do not have permission to edit this plugin")))

        if not obj.placeholder.check_source(request.user):
            message = force_str(_("You do not have permission to edit this plugin"))
            return HttpResponseForbidden(message)

        response = plugin_instance.change_view(request, str(plugin_id))

        plugin = getattr(plugin_instance, 'saved_object', None)

        if plugin:
            plugin.placeholder.clear_cache(plugin.language)

        if plugin_instance._operation_token:
            self._send_post_placeholder_operation(
                request,
                operation=operations.CHANGE_PLUGIN,
                token=plugin_instance._operation_token,
                old_plugin=obj,
                new_plugin=plugin,
                placeholder=plugin.placeholder,
            )
        return response

    @method_decorator(require_POST)
    @xframe_options_sameorigin
    @transaction.atomic
    def move_plugin(self, request):
        """
        Performs a move or a "paste" operation (when «move_a_copy» is set)

        POST request with following parameters:
        - plugin_id
        - placeholder_id
        - plugin_language (optional)
        - plugin_parent (optional)
        - plugin_order (array, optional)
        - move_a_copy (Boolean, optional) (anything supplied here except a case-
                                        insensitive "false" is True)
        NOTE: If move_a_copy is set, the plugin_order should contain an item
              '__COPY__' with the desired destination of the copied plugin.
        """
        # plugin_id and placeholder_id are required, so, if nothing is supplied,
        # an ValueError exception will be raised by get_int().
        try:
            plugin_id = get_int(request.POST.get('plugin_id'))
        except TypeError:
            raise RuntimeError("'plugin_id' is a required parameter.")

        plugin = self._get_plugin_from_id(plugin_id)

        if request.POST.get('placeholder_id'):
            placeholder_id = get_int(request.POST.get('placeholder_id'))
            placeholder = Placeholder.objects.get(pk=placeholder_id)
            assert plugin.placeholder_id != placeholder_id
        else:
            placeholder = None

        # The rest are optional
        parent_id = get_int(request.POST.get('plugin_parent', ""), None)
        target_language = request.POST['target_language']
        move_a_copy = request.POST.get('move_a_copy')
        move_a_copy = (
            move_a_copy and move_a_copy != "0" and move_a_copy.lower() != "false"
        )
        move_to_clipboard = placeholder == request.toolbar.clipboard
        source_placeholder = plugin.placeholder

        if placeholder and placeholder != source_placeholder:
            try:
                template = self.get_placeholder_template(request, placeholder)
                has_reached_plugin_limit(placeholder, plugin.plugin_type,
                                         target_language, template=template)
            except PluginLimitReached as er:
                return HttpResponseBadRequest(er)

        # True if the plugin is not being moved from the clipboard
        # to a placeholder or from a placeholder to the clipboard.
        move_a_plugin = not move_a_copy and not move_to_clipboard

        if parent_id and plugin.parent_id != parent_id:
            target_pl = placeholder or plugin.placeholder

            if move_a_plugin:
                target_parent = get_object_or_404(
                    CMSPlugin,
                    pk=parent_id,
                    language=target_language,
                    placeholder=target_pl,
                )
            else:
                target_parent = get_object_or_404(CMSPlugin, pk=parent_id)
        elif parent_id:
            target_parent = plugin.parent
        else:
            target_parent = None

        new_plugin = None
        fetch_tree = False

        if move_a_copy and plugin.plugin_type == "PlaceholderPlugin":
            new_plugins = self._paste_placeholder(
                request,
                plugin=plugin,
                target_language=target_language,
                target_placeholder=placeholder,
                target_position=int(request.POST['target_position']),
            )
        elif move_a_copy:
            fetch_tree = True
            new_plugin = self._paste_plugin(
                request,
                plugin=plugin,
                target_parent=target_parent,
                target_language=target_language,
                target_placeholder=placeholder,
                target_position=int(request.POST['target_position']),
            )
        elif move_to_clipboard:
            new_plugin = self._cut_plugin(
                request,
                plugin=plugin,
                target_language=target_language,
                target_placeholder=placeholder,
            )
            new_plugins = [new_plugin]
        else:
            fetch_tree = True
            new_plugin = self._move_plugin(
                request,
                plugin=plugin,
                target_parent=target_parent,
                target_position=int(request.POST['target_position']),
                target_placeholder=placeholder,
            )

        if new_plugin and fetch_tree:
            root = (new_plugin.parent or new_plugin)
            new_plugins = [root] + list(root.get_descendants())
        data = get_plugin_tree_as_json(request, new_plugins)
        return HttpResponse(data, content_type='application/json')

    def _paste_plugin(self, request, plugin, target_language,
                      target_placeholder, target_position, target_parent=None):
        plugins = [plugin] + list(plugin.get_descendants())

        if not self.has_copy_from_clipboard_permission(request, target_placeholder, plugins):
            message = force_str(_("You have no permission to paste this plugin"))
            raise PermissionDenied(message)

        if not target_placeholder.check_source(request.user):
            message = force_str(_("You have no permission to paste this plugin"))
            raise PermissionDenied(message)

        if target_parent:
            target_parent_id = target_parent.pk
        else:
            target_parent_id = None

        action_token = self._send_pre_placeholder_operation(
            request,
            operation=operations.PASTE_PLUGIN,
            plugin=plugin,
            target_language=target_language,
            target_placeholder=target_placeholder,
            target_parent_id=target_parent_id,
        )

        target_last_plugin = target_placeholder.get_last_plugin(plugin.language)

        if target_last_plugin:
            target_offset = target_last_plugin.position + len(plugins)
            target_placeholder._shift_plugin_positions(
                target_language,
                start=target_position,
                offset=target_offset,
            )

        new_plugins = copy_plugins_to_placeholder(
            plugins,
            placeholder=target_placeholder,
            language=target_language,
            root_plugin=target_parent,
            start_positions={target_language: target_position},
        )
        root_plugin = new_plugins[0]
        target_placeholder.clear_cache(target_language)

        # Fetch from db to update position and other tree values
        root_plugin.refresh_from_db()

        self._send_post_placeholder_operation(
            request,
            operation=operations.PASTE_PLUGIN,
            plugin=root_plugin.get_bound_plugin(),
            token=action_token,
            target_language=target_language,
            target_placeholder=target_placeholder,
            target_parent_id=target_parent_id,
        )
        return root_plugin

    def _paste_placeholder(self, request, plugin, target_language,
                           target_placeholder, target_position):
        plugins = plugin.placeholder_ref.get_plugins_list()

        if not self.has_copy_from_clipboard_permission(request, target_placeholder, plugins):
            message = force_str(_("You have no permission to paste this placeholder"))
            raise PermissionDenied(message)

        if not target_placeholder.check_source(request.user):
            message = force_str(_("You have no permission to paste this placeholder"))
            raise PermissionDenied(message)

        action_token = self._send_pre_placeholder_operation(
            request,
            operation=operations.PASTE_PLACEHOLDER,
            plugins=plugins,
            target_language=target_language,
            target_placeholder=target_placeholder,
        )

        target_last_plugin = target_placeholder.get_last_plugin(plugin.language)

        if target_last_plugin:
            target_offset = target_last_plugin.position + len(plugins)
            target_placeholder._shift_plugin_positions(
                target_language,
                start=target_position,
                offset=target_offset,
            )

        new_plugins = copy_plugins_to_placeholder(
            plugins,
            placeholder=target_placeholder,
            language=target_language,
            start_positions={target_language: target_position},
        )
        new_plugin_ids = (new.pk for new in new_plugins)
        target_placeholder.clear_cache(plugin.language)

        new_plugins = (
            CMSPlugin
            .objects
            .filter(pk__in=new_plugin_ids)
            .select_related('placeholder')
        )
        new_plugins = list(new_plugins)

        self._send_post_placeholder_operation(
            request,
            operation=operations.PASTE_PLACEHOLDER,
            token=action_token,
            plugins=new_plugins,
            target_language=target_language,
            target_placeholder=target_placeholder,
        )
        return new_plugins

    def _move_plugin(self, request, plugin, target_position, target_placeholder=None, target_parent=None):
        language = plugin.language
        source_placeholder = plugin.placeholder

        if not self.has_move_plugin_permission(request, plugin, source_placeholder):
            message = force_str(_("You have no permission to move this plugin"))
            raise PermissionDenied(message)

        if target_placeholder and not self.has_move_plugin_permission(request, plugin, target_placeholder):
            message = force_str(_("You have no permission to move this plugin"))
            raise PermissionDenied(message)

        if not source_placeholder.check_source(request.user):
            message = force_str(_("You have no permission to move this plugin"))
            raise PermissionDenied(message)

        if target_placeholder and not target_placeholder.check_source(request.user):
            message = force_str(_("You have no permission to move this plugin"))
            raise PermissionDenied(message)

        if target_parent:
            target_parent_id = target_parent.pk
        else:
            target_parent_id = None

        action_token = self._send_pre_placeholder_operation(
            request,
            operation=operations.MOVE_PLUGIN,
            plugin=plugin,
            source_language=plugin.language,
            source_placeholder=source_placeholder,
            source_parent_id=plugin.parent_id,
            target_language=language,
            target_placeholder=(target_placeholder or source_placeholder),
            target_parent_id=target_parent_id,
        )

        source_placeholder.move_plugin(
            plugin=plugin,
            target_position=target_position,
            target_placeholder=target_placeholder,
            target_plugin=target_parent,
        )

        # Refresh plugin to get new position values
        updated_plugin = plugin.reload()

        if target_placeholder:
            target_placeholder.clear_cache(language)
        source_placeholder.clear_cache(language)

        self._send_post_placeholder_operation(
            request,
            operation=operations.MOVE_PLUGIN,
            plugin=updated_plugin.get_bound_plugin(),
            token=action_token,
            source_language=language,
            source_placeholder=source_placeholder,
            source_parent_id=updated_plugin.parent_id,
            target_language=language,
            target_placeholder=(target_placeholder or source_placeholder),
            target_parent_id=target_parent_id,
        )
        return updated_plugin

    def _cut_plugin(self, request, plugin, target_language, target_placeholder):
        source_placeholder = plugin.placeholder

        if not self.has_move_plugin_permission(request, plugin, source_placeholder):
            message = force_str(_("You have no permission to cut this plugin"))
            raise PermissionDenied(message)

        if not source_placeholder.check_source(request.user):
            message = force_str(_("You have no permission to cut this plugin"))
            raise PermissionDenied(message)

        action_token = self._send_pre_placeholder_operation(
            request,
            operation=operations.CUT_PLUGIN,
            plugin=plugin,
            clipboard=target_placeholder,
            clipboard_language=target_language,
            source_language=plugin.language,
            source_placeholder=source_placeholder,
            source_parent_id=plugin.parent_id,
            source_order=[],
        )

        # Empty the clipboard
        target_placeholder.clear()
        source_placeholder.move_plugin(
            plugin,
            target_position=1,
            target_placeholder=target_placeholder,
            target_plugin=None,
        )
        source_placeholder.clear_cache(plugin.language)
        updated_plugin = plugin.reload()

        self._send_post_placeholder_operation(
            request,
            operation=operations.CUT_PLUGIN,
            token=action_token,
            plugin=updated_plugin.get_bound_plugin(),
            clipboard=target_placeholder,
            clipboard_language=target_language,
            source_language=plugin.language,
            source_placeholder=source_placeholder,
            source_parent_id=plugin.parent_id,
            source_order=[],
        )
        return updated_plugin

    @xframe_options_sameorigin
    def delete_plugin(self, request, plugin_id):
        plugin = self._get_plugin_from_id(plugin_id)

        if not self.has_delete_plugin_permission(request, plugin):
            return HttpResponseForbidden(force_str(
                _("You do not have permission to delete this plugin")))

        if not plugin.placeholder.check_source(request.user):
            message = force_str(_("You do not have permission to delete this plugin"))
            raise PermissionDenied(message)

        opts = plugin._meta
        get_deleted_objects_additional_kwargs = {'request': request}
        deleted_objects, __, perms_needed, protected = get_deleted_objects(
            [plugin], admin_site=self.admin_site,
            **get_deleted_objects_additional_kwargs
        )

        if request.POST:  # The user has already confirmed the deletion.
            if perms_needed:
                raise PermissionDenied(_("You do not have permission to delete this plugin"))
            obj_display = force_str(plugin)
            placeholder = plugin.placeholder
            plugin_tree_order = placeholder.get_plugin_tree_order(
                language=plugin.language,
                parent_id=plugin.parent_id,
            )

            operation_token = self._send_pre_placeholder_operation(
                request,
                operation=operations.DELETE_PLUGIN,
                plugin=plugin,
                placeholder=placeholder,
                tree_order=plugin_tree_order,
            )
            placeholder.delete_plugin(plugin)
            placeholder.clear_cache(plugin.language)

            self.message_user(request, _('The %(name)s plugin "%(obj)s" was deleted successfully.') % {
                'name': force_str(opts.verbose_name), 'obj': force_str(obj_display)})

            self._send_post_placeholder_operation(
                request,
                operation=operations.DELETE_PLUGIN,
                token=operation_token,
                plugin=plugin,
                placeholder=placeholder,
                tree_order=plugin_tree_order,
            )
            return HttpResponseRedirect(admin_reverse('index', current_app=self.admin_site.name))

        plugin_name = force_str(plugin.get_plugin_class().name)

        if perms_needed or protected:
            title = _("Cannot delete %(name)s") % {"name": plugin_name}
        else:
            title = _("Are you sure?")
        context = {
            "title": title,
            "object_name": plugin_name,
            "object": plugin,
            "deleted_objects": deleted_objects,
            "perms_lacking": perms_needed,
            "protected": protected,
            "opts": opts,
            "app_label": opts.app_label,
        }
        request.current_app = self.admin_site.name
        return TemplateResponse(
            request, "admin/cms/page/plugin/delete_confirmation.html", context
        )

    @xframe_options_sameorigin
    def clear_placeholder(self, request, placeholder_id):
        placeholder = get_object_or_404(Placeholder, pk=placeholder_id)
        language = request.GET.get('language')

        if placeholder.pk == request.toolbar.clipboard.pk:
            # User is clearing the clipboard, no need for permission
            # checks here as the clipboard is unique per user.
            # There could be a case where a plugin has relationship to
            # an object the user does not have permission to delete.
            placeholder.clear(language)
            return HttpResponseRedirect(admin_reverse('index', current_app=self.admin_site.name))

        if not self.has_clear_placeholder_permission(request, placeholder, language):
            return HttpResponseForbidden(force_str(_("You do not have permission to clear this placeholder")))

        if not placeholder.check_source(request.user):
            message = force_str(_("You do not have permission to clear this placeholder"))
            raise PermissionDenied(message)

        opts = Placeholder._meta
        plugins = placeholder.get_plugins_list(language)

        get_deleted_objects_additional_kwargs = {'request': request}
        deleted_objects, __, perms_needed, protected = get_deleted_objects(
            plugins, admin_site=self.admin_site,
            **get_deleted_objects_additional_kwargs
        )

        obj_display = force_str(placeholder)

        if request.POST:
            # The user has already confirmed the deletion.
            if perms_needed:
                return HttpResponseForbidden(force_str(_("You do not have permission to clear this placeholder")))

            operation_token = self._send_pre_placeholder_operation(
                request,
                operation=operations.CLEAR_PLACEHOLDER,
                plugins=plugins,
                placeholder=placeholder,
            )

            placeholder.clear(language)
            placeholder.clear_cache(language)

            self.message_user(request, _('The placeholder "%(obj)s" was cleared successfully.') % {
                'obj': obj_display})

            self._send_post_placeholder_operation(
                request,
                operation=operations.CLEAR_PLACEHOLDER,
                token=operation_token,
                plugins=plugins,
                placeholder=placeholder,
            )
            return HttpResponseRedirect(admin_reverse('index', current_app=self.admin_site.name))

        if perms_needed or protected:
            title = _("Cannot delete %(name)s") % {"name": obj_display}
        else:
            title = _("Are you sure?")

        context = {
            "title": title,
            "object_name": _("placeholder"),
            "object": placeholder,
            "deleted_objects": deleted_objects,
            "perms_lacking": perms_needed,
            "protected": protected,
            "opts": opts,
            "app_label": opts.app_label,
        }
        request.current_app = self.admin_site.name
        return TemplateResponse(request, "admin/cms/page/plugin/delete_confirmation.html", context)


admin.site.register(Placeholder, PlaceholderAdmin)



from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

from cms.app_registration import (
    autodiscover_cms_configs,
    backwards_compatibility_config,
    configure_cms_apps,
    get_cms_extension_apps,
    ready_cms_apps,
)
from cms.utils.compat.dj import is_installed as app_is_installed


def validate_dependencies():
    """
    Check for installed apps, their versions and configuration options
    """
    if not app_is_installed('treebeard'):
        raise ImproperlyConfigured(
            'django CMS requires django-treebeard. Please install it and add "treebeard" to INSTALLED_APPS.'
        )


def validate_settings():
    """
    Check project settings file for required options
    """
    try:
        django_backend = [x for x in settings.TEMPLATES
                          if x['BACKEND'] == 'django.template.backends.django.DjangoTemplates'][0]
    except IndexError:
        raise ImproperlyConfigured("django CMS requires django.template.context_processors.request in "
                                   "'django.template.backends.django.DjangoTemplates' context processors.")

    context_processors = django_backend.get('OPTIONS', {}).get('context_processors', [])
    core_request = 'django.core.context_processors.request'
    template_request = 'django.template.context_processors.request'
    if core_request not in context_processors and template_request not in context_processors:
        raise ImproperlyConfigured("django CMS requires django.template.context_processors.request in "
                                   "'django.template.backends.django.DjangoTemplates' context processors.")


def setup():
    """
    Gather all checks and validations
    """
    from cms.plugin_pool import plugin_pool
    validate_dependencies()
    validate_settings()
    plugin_pool.validate_templates()


def setup_cms_apps():
    """
    Check for django apps which provide functionality that extends the
    cms. Configure all apps which have configs that declare use of
    any of this functionality.
    """
    autodiscover_cms_configs()
    cms_apps = get_cms_extension_apps()
    configure_cms_apps(cms_apps)
    backwards_compatibility_config()
    ready_cms_apps(cms_apps)
